import warnings
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ...data.datasets import BaseDataset
from ..base.base_utils import ModelOutput
from ..nn import BaseDecoder, BaseEncoder
from ..nn.default_architectures import Encoder_SVAE_MLP, Encoder_VAE_MLP
from ..vae import VAE
from .pvae_config import PoincareVAEConfig
from .pvae_utils import PoincareBall, RiemannianNormal, WrappedNormal


class PoincareVAE(VAE):
    """Poincaré Variational Autoencoder model.

    Args:
        model_config (PoincareVAEConfig): The Poincaré Variational Autoencoder configuration
            setting the main parameters of the model.

        encoder (BaseEncoder): An instance of BaseEncoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        decoder (BaseDecoder): An instance of BaseDecoder (inheriting from `torch.nn.Module` which
            plays the role of decoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

    .. note::
        For high dimensional data we advice you to provide you own network architectures. With the
        provided MLP you may end up with a ``MemoryError``.
    """

    def __init__(
        self,
        model_config: PoincareVAEConfig,
        encoder: Optional[BaseEncoder] = None,
        decoder: Optional[BaseDecoder] = None,
    ):

        VAE.__init__(self, model_config=model_config, encoder=encoder, decoder=decoder)

        self.model_name = "PoincareVAE"

        self.latent_manifold = PoincareBall(
            dim=model_config.latent_dim, c=model_config.curvature
        )

        if model_config.prior_distribution == "riemannian_normal":
            self.prior = RiemannianNormal
        else:
            self.prior = WrappedNormal

        if model_config.posterior_distribution == "riemannian_normal":
            warnings.warn(
                "Carefull, this model expects the encoder to give a one dimensional "
                "`log_concentration` tensor for the Riemannian normal distribution. "
                "Make sure the encoder actually outputs this."
            )
            self.posterior = RiemannianNormal
        else:
            self.posterior = WrappedNormal

        if encoder is None:
            if model_config.posterior_distribution == "riemannian_normal":
                encoder = Encoder_SVAE_MLP(model_config)
            else:
                encoder = Encoder_VAE_MLP(model_config)
            self.model_config.uses_default_encoder = True

        else:
            self.model_config.uses_default_encoder = False

        self.set_encoder(encoder)

        self._pz_mu = nn.Parameter(
            torch.zeros(1, model_config.latent_dim), requires_grad=False
        )
        self._pz_logvar = nn.Parameter(torch.zeros(1, 1), requires_grad=False)

    def forward(self, inputs: BaseDataset, **kwargs):
        """
        The VAE model

        Args:
            inputs (BaseDataset): The training dataset with labels

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters

        """

        x = inputs["data"]

        encoder_output = self.encoder(x)

        if self.model_config.posterior_distribution == "riemannian_normal":
            mu, log_var = encoder_output.embedding, encoder_output.log_concentration
        else:
            mu, log_var = encoder_output.embedding, encoder_output.log_covariance

        std = torch.exp(0.5 * log_var)

        qz_x = self.posterior(loc=mu, scale=std, manifold=self.latent_manifold)
        z = qz_x.rsample(torch.Size([1]))

        recon_x = self.decoder(z.squeeze(0))["reconstruction"]

        loss, recon_loss, kld = self.loss_function(recon_x, x, z, qz_x)

        output = ModelOutput(
            recon_loss=recon_loss,
            reg_loss=kld,
            loss=loss,
            recon_x=recon_x,
            z=z.squeeze(0),
        )

        return output

    def loss_function(self, recon_x, x, z, qz_x):

        if self.model_config.reconstruction_loss == "mse":

            recon_loss = F.mse_loss(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        elif self.model_config.reconstruction_loss == "bce":

            recon_loss = F.binary_cross_entropy(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        pz = self.prior(
            loc=self._pz_mu, scale=self._pz_logvar.exp(), manifold=self.latent_manifold
        )

        KLD = (qz_x.log_prob(z) - pz.log_prob(z)).sum(-1).squeeze(0)

        return (recon_loss + KLD).mean(dim=0), recon_loss.mean(dim=0), KLD.mean(dim=0)

    def interpolate(
        self,
        starting_inputs: torch.Tensor,
        ending_inputs: torch.Tensor,
        granularity: int = 10,
    ):
        """This function performs a geodesic interpolation in the poincaré disk of the autoencoder
        from starting inputs to ending inputs. It returns the interpolation trajectories.

        Args:
            starting_inputs (torch.Tensor): The starting inputs in the interpolation of shape
                [B x input_dim]
            ending_inputs (torch.Tensor): The starting inputs in the interpolation of shape
                [B x input_dim]
            granularity (int): The granularity of the interpolation.

        Returns:
            torch.Tensor: A tensor of shape [B x granularity x input_dim] containing the
            interpolation trajectories.
        """
        assert starting_inputs.shape[0] == ending_inputs.shape[0], (
            "The number of starting_inputs should equal the number of ending_inputs. Got "
            f"{starting_inputs.shape[0]} sampler for starting_inputs and {ending_inputs.shape[0]} "
            "for endinging_inputs."
        )

        starting_z = self.encoder(starting_inputs).embedding
        ending_z = self.encoder(ending_inputs).embedding
        t = torch.linspace(0, 1, granularity).to(starting_inputs.device)

        inter_geo = torch.zeros(
            starting_inputs.shape[0], granularity, starting_z.shape[-1]
        ).to(starting_z.device)

        for i, t_i in enumerate(t):
            z_i = self.latent_manifold.geodesic(t_i, starting_z, ending_z)
            inter_geo[:, i, :] = z_i

        decoded_geo = self.decoder(
            inter_geo.reshape(
                (starting_z.shape[0] * t.shape[0],) + (starting_z.shape[1:])
            )
        ).reconstruction.reshape(
            (
                starting_inputs.shape[0],
                t.shape[0],
            )
            + (starting_inputs.shape[1:])
        )
        return decoded_geo

    def get_nll(self, data, n_samples=1, batch_size=100):
        """
        Function computed the estimate negative log-likelihood of the model. It uses importance
        sampling method with the approximate posterior distribution. This may take a while.

        Args:
            data (torch.Tensor): The input data from which the log-likelihood should be estimated.
                Data must be of shape [Batch x n_channels x ...]
            n_samples (int): The number of importance samples to use for estimation
            batch_size (int): The batchsize to use to avoid memory issues
        """

        if n_samples <= batch_size:
            n_full_batch = 1
        else:
            n_full_batch = n_samples // batch_size
            n_samples = batch_size

        log_p = []

        for i in range(len(data)):
            x = data[i].unsqueeze(0)

            log_p_x = []

            for j in range(n_full_batch):

                x_rep = torch.cat(batch_size * [x])

                encoder_output = self.encoder(x_rep)
                if self.model_config.posterior_distribution == "riemannian_normal":
                    mu, log_var = (
                        encoder_output.embedding,
                        encoder_output.log_concentration,
                    )
                else:
                    mu, log_var = (
                        encoder_output.embedding,
                        encoder_output.log_covariance,
                    )

                std = torch.exp(0.5 * log_var)

                qz_x = self.posterior(loc=mu, scale=std, manifold=self.latent_manifold)
                z = qz_x.rsample(torch.Size([1]))

                pz = self.prior(
                    loc=self._pz_mu,
                    scale=self._pz_logvar.exp(),
                    manifold=self.latent_manifold,
                )

                log_q_z_given_x = qz_x.log_prob(z).sum(-1).squeeze(0)
                log_p_z = pz.log_prob(z).sum(-1).squeeze(0)

                recon_x = self.decoder(z.squeeze(0))["reconstruction"]

                if self.model_config.reconstruction_loss == "mse":

                    log_p_x_given_z = -0.5 * F.mse_loss(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1) - torch.tensor(
                        [np.prod(self.input_dim) / 2 * np.log(np.pi * 2)]
                    ).to(
                        data.device
                    )  # decoding distribution is assumed unit variance  N(mu, I)

                elif self.model_config.reconstruction_loss == "bce":

                    log_p_x_given_z = -F.binary_cross_entropy(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1)

                log_p_x.append(log_p_x_given_z + log_p_z - log_q_z_given_x)

            log_p_x = torch.cat(log_p_x)

            log_p.append((torch.logsumexp(log_p_x, 0) - np.log(len(log_p_x))).item())
        return np.mean(log_p)



import datetime
import logging
import os
from copy import deepcopy
from typing import List, Optional

import torch
import torch.distributed as dist
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler

from ...data.datasets import BaseDataset
from ...models import BaseAE
from ..base_trainer import BaseTrainer
from ..trainer_utils import set_seed
from ..training_callbacks import TrainingCallback
from .coupled_optimizer_trainer_config import CoupledOptimizerTrainerConfig

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class CoupledOptimizerTrainer(BaseTrainer):
    """Trainer using distinct optimizers for encoder and decoder nn.

    Args:
        model (BaseAE): The model to train

        train_dataset (BaseDataset): The training dataset of type
            :class:`~XGen.data.dataset.BaseDataset`

        training_args (CoupledOptimizerTrainerConfig): The training arguments summarizing the main
            parameters used for training. If None, a basic training instance of
            :class:`CoupledOptimizerTrainerConfig` is used. Default: None.

        encoder_optimizer (~torch.optim.Optimizer): An instance of `torch.optim.Optimizer` used for
            training the encoder. If None, a :class:`~torch.optim.Adam` optimizer is used.
            Default: None.

        decoder_optimizer (~torch.optim.Optimizer): An instance of `torch.optim.Optimizer` used for
            training the decoder. If None, a :class:`~torch.optim.Adam` optimizer is used.
            Default: None.
    """

    def __init__(
        self,
        model: BaseAE,
        train_dataset: BaseDataset,
        eval_dataset: Optional[BaseDataset] = None,
        training_config: Optional[CoupledOptimizerTrainerConfig] = None,
        callbacks: List[TrainingCallback] = None,
    ):

        BaseTrainer.__init__(
            self,
            model=model,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            training_config=training_config,
            callbacks=callbacks,
        )

    def set_encoder_optimizer(self):
        encoder_optimizer_cls = getattr(
            optim, self.training_config.encoder_optimizer_cls
        )

        if self.training_config.encoder_optimizer_params is not None:
            if self.distributed:
                encoder_optimizer = encoder_optimizer_cls(
                    self.model.module.encoder.parameters(),
                    lr=self.training_config.encoder_learning_rate,
                    **self.training_config.encoder_optimizer_params,
                )
            else:
                encoder_optimizer = encoder_optimizer_cls(
                    self.model.encoder.parameters(),
                    lr=self.training_config.encoder_learning_rate,
                    **self.training_config.encoder_optimizer_params,
                )
        else:
            if self.distributed:
                encoder_optimizer = encoder_optimizer_cls(
                    self.model.module.encoder.parameters(),
                    lr=self.training_config.encoder_learning_rate,
                )
            else:
                encoder_optimizer = encoder_optimizer_cls(
                    self.model.encoder.parameters(),
                    lr=self.training_config.encoder_learning_rate,
                )

        self.encoder_optimizer = encoder_optimizer

    def set_encoder_scheduler(self):
        if self.training_config.encoder_scheduler_cls is not None:
            encoder_scheduler_cls = getattr(
                lr_scheduler, self.training_config.encoder_scheduler_cls
            )

            if self.training_config.encoder_scheduler_params is not None:
                scheduler = encoder_scheduler_cls(
                    self.encoder_optimizer,
                    **self.training_config.encoder_scheduler_params,
                )
            else:
                scheduler = encoder_scheduler_cls(self.encoder_optimizer)

        else:
            scheduler = None

        self.encoder_scheduler = scheduler

    def set_decoder_optimizer(self):
        decoder_cls = getattr(optim, self.training_config.decoder_optimizer_cls)

        if self.training_config.decoder_optimizer_params is not None:
            if self.distributed:
                decoder_optimizer = decoder_cls(
                    self.model.module.decoder.parameters(),
                    lr=self.training_config.decoder_learning_rate,
                    **self.training_config.decoder_optimizer_params,
                )
            else:
                decoder_optimizer = decoder_cls(
                    self.model.decoder.parameters(),
                    lr=self.training_config.decoder_learning_rate,
                    **self.training_config.decoder_optimizer_params,
                )

        else:
            if self.distributed:
                decoder_optimizer = decoder_cls(
                    self.model.module.decoder.parameters(),
                    lr=self.training_config.decoder_learning_rate,
                )
            else:
                decoder_optimizer = decoder_cls(
                    self.model.decoder.parameters(),
                    lr=self.training_config.decoder_learning_rate,
                )

        self.decoder_optimizer = decoder_optimizer

    def set_decoder_scheduler(self) -> torch.optim.lr_scheduler:
        if self.training_config.decoder_scheduler_cls is not None:
            decoder_scheduler_cls = getattr(
                lr_scheduler, self.training_config.decoder_scheduler_cls
            )

            if self.training_config.decoder_scheduler_params is not None:
                scheduler = decoder_scheduler_cls(
                    self.decoder_optimizer,
                    **self.training_config.decoder_scheduler_params,
                )
            else:
                scheduler = decoder_scheduler_cls(self.decoder_optimizer)

        else:
            scheduler = None

        self.decoder_scheduler = scheduler

    def _optimizers_step(self, model_output):

        encoder_loss = model_output.encoder_loss
        decoder_loss = model_output.decoder_loss

        # Reset optimizers
        if model_output.update_encoder:
            self.encoder_optimizer.zero_grad()
            encoder_loss.backward(retain_graph=True)

        if model_output.update_decoder:
            self.decoder_optimizer.zero_grad()
            decoder_loss.backward(retain_graph=True)

        if model_output.update_encoder:
            self.encoder_optimizer.step()

        if model_output.update_decoder:
            self.decoder_optimizer.step()

    def _schedulers_step(self, encoder_metrics=None, decoder_metrics=None):

        if self.encoder_scheduler is None:
            pass

        elif isinstance(self.encoder_scheduler, lr_scheduler.ReduceLROnPlateau):
            self.encoder_scheduler.step(encoder_metrics)

        else:
            self.encoder_scheduler.step()

        if self.decoder_scheduler is None:
            pass

        elif isinstance(self.decoder_scheduler, lr_scheduler.ReduceLROnPlateau):
            self.decoder_scheduler.step(decoder_metrics)

        else:
            self.decoder_scheduler.step()

    def prepare_training(self):

        # set random seed
        set_seed(self.training_config.seed)

        # set autoencoder optimizer and scheduler
        self.set_encoder_optimizer()
        self.set_encoder_scheduler()

        # set discriminator optimizer and scheduler
        self.set_decoder_optimizer()
        self.set_decoder_scheduler()

        # create foder for saving
        self._set_output_dir()

        # set callbacks
        self._setup_callbacks()

    def train(self, log_output_dir: str = None):
        """This function is the main training function

        Args:
            log_output_dir (str): The path in which the log will be stored
        """

        self.prepare_training()

        self.callback_handler.on_train_begin(
            training_config=self.training_config, model_config=self.model_config
        )

        log_verbose = False

        msg = (
            f"Training params:\n - max_epochs: {self.training_config.num_epochs}\n"
            " - per_device_train_batch_size: "
            f"{self.training_config.per_device_train_batch_size}\n"
            " - per_device_eval_batch_size: "
            f"{self.training_config.per_device_eval_batch_size}\n"
            f" - checkpoint saving every: {self.training_config.steps_saving}\n"
            f"Encoder Optimizer: {self.encoder_optimizer}\n"
            f"Encoder Scheduler: {self.encoder_scheduler}\n"
            f"Decoder Optimizer: {self.decoder_optimizer}\n"
            f"Decoder Scheduler: {self.decoder_scheduler}\n"
        )

        if self.is_main_process:
            logger.info(msg)

        # set up log file
        if log_output_dir is not None and self.is_main_process:
            log_verbose = True

            file_logger = self._get_file_logger(log_output_dir=log_output_dir)

            file_logger.info(msg)

        if self.is_main_process:
            logger.info("Successfully launched training !\n")

        # set best losses for early stopping
        best_train_loss = 1e10
        best_eval_loss = 1e10

        for epoch in range(1, self.training_config.num_epochs + 1):

            self.callback_handler.on_epoch_begin(
                training_config=self.training_config,
                epoch=epoch,
                train_loader=self.train_loader,
                eval_loader=self.eval_loader,
            )

            metrics = {}

            train_losses = self.train_step(epoch)

            [
                epoch_train_loss,
                epoch_train_encoder_loss,
                epoch_train_decoder_loss,
            ] = train_losses
            metrics["train_epoch_loss"] = epoch_train_loss
            metrics["train_encoder_loss"] = epoch_train_encoder_loss
            metrics["train_decoder_loss"] = epoch_train_decoder_loss

            if self.eval_dataset is not None:
                eval_losses = self.eval_step(epoch)

                [
                    epoch_eval_loss,
                    epoch_eval_encoder_loss,
                    epoch_eval_decoder_loss,
                ] = eval_losses
                metrics["eval_epoch_loss"] = epoch_eval_loss
                metrics["eval_encoder_loss"] = epoch_eval_encoder_loss
                metrics["eval_decoder_loss"] = epoch_eval_decoder_loss

                self._schedulers_step(
                    encoder_metrics=epoch_eval_encoder_loss,
                    decoder_metrics=epoch_eval_decoder_loss,
                )

            else:
                epoch_eval_loss = best_eval_loss
                self._schedulers_step(
                    encoder_metrics=epoch_train_encoder_loss,
                    decoder_metrics=epoch_train_decoder_loss,
                )

            if (
                epoch_eval_loss < best_eval_loss
                and not self.training_config.keep_best_on_train
            ):
                best_eval_loss = epoch_eval_loss
                best_model = deepcopy(self.model)
                self._best_model = best_model

            elif (
                epoch_train_loss < best_train_loss
                and self.training_config.keep_best_on_train
            ):
                best_train_loss = epoch_train_loss
                best_model = deepcopy(self.model)
                self._best_model = best_model

            if (
                self.training_config.steps_predict is not None
                and epoch % self.training_config.steps_predict == 0
                and self.is_main_process
            ):

                true_data, reconstructions, generations = self.predict(best_model)

                self.callback_handler.on_prediction_step(
                    self.training_config,
                    true_data=true_data,
                    reconstructions=reconstructions,
                    generations=generations,
                    global_step=epoch,
                )

            self.callback_handler.on_epoch_end(training_config=self.training_config)

            # save checkpoints
            if (
                self.training_config.steps_saving is not None
                and epoch % self.training_config.steps_saving == 0
            ):
                if self.is_main_process:
                    self.save_checkpoint(
                        model=best_model, dir_path=self.training_dir, epoch=epoch
                    )
                    logger.info(f"Saved checkpoint at epoch {epoch}\n")

                    if log_verbose:
                        file_logger.info(f"Saved checkpoint at epoch {epoch}\n")

            self.callback_handler.on_log(
                self.training_config,
                metrics,
                logger=logger,
                global_step=epoch,
                rank=self.rank,
            )

        final_dir = os.path.join(self.training_dir, "final_model")

        if self.is_main_process:
            self.save_model(best_model, dir_path=final_dir)
            logger.info("----------------------------------")
            logger.info("Training ended!")
            logger.info(f"Saved final model in {final_dir}")

        if self.distributed:
            dist.destroy_process_group()

        self.callback_handler.on_train_end(self.training_config)

    def eval_step(self, epoch: int):
        """Perform an evaluation step

        Parameters:
            epoch (int): The current epoch number

        Returns:
            (torch.Tensor): The evaluation loss
        """
        self.callback_handler.on_eval_step_begin(
            training_config=self.training_config,
            eval_loader=self.eval_loader,
            epoch=epoch,
            rank=self.rank,
        )

        self.model.eval()

        epoch_encoder_loss = 0
        epoch_decoder_loss = 0
        epoch_loss = 0

        for inputs in self.eval_loader:

            inputs = self._set_inputs_to_device(inputs)

            try:
                with torch.no_grad():

                    model_output = self.model(
                        inputs,
                        epoch=epoch,
                        dataset_size=len(self.eval_loader.dataset),
                        uses_ddp=self.distributed,
                    )

            except RuntimeError:
                model_output = self.model(
                    inputs,
                    epoch=epoch,
                    dataset_size=len(self.eval_loader.dataset),
                    uses_ddp=self.distributed,
                )

            encoder_loss = model_output.encoder_loss
            decoder_loss = model_output.decoder_loss

            loss = encoder_loss + decoder_loss

            epoch_encoder_loss += encoder_loss.item()
            epoch_decoder_loss += decoder_loss.item()
            epoch_loss += loss.item()

            if epoch_loss != epoch_loss:
                raise ArithmeticError("NaN detected in eval loss")

            self.callback_handler.on_eval_step_end(training_config=self.training_config)

        epoch_encoder_loss /= len(self.eval_loader)
        epoch_decoder_loss /= len(self.eval_loader)
        epoch_loss /= len(self.eval_loader)

        return (epoch_loss, epoch_encoder_loss, epoch_decoder_loss)

    def train_step(self, epoch: int):
        """The trainer performs training loop over the train_loader.

        Parameters:
            epoch (int): The current epoch number

        Returns:
            (torch.Tensor): The step training loss
        """
        self.callback_handler.on_train_step_begin(
            training_config=self.training_config,
            train_loader=self.train_loader,
            epoch=epoch,
            rank=self.rank,
        )

        # set model in train model
        self.model.train()

        epoch_encoder_loss = 0
        epoch_decoder_loss = 0
        epoch_loss = 0

        for inputs in self.train_loader:

            inputs = self._set_inputs_to_device(inputs)

            model_output = self.model(
                inputs,
                epoch=epoch,
                dataset_size=len(self.train_loader.dataset),
                uses_ddp=self.distributed,
            )

            self._optimizers_step(model_output)

            encoder_loss = model_output.encoder_loss
            decoder_loss = model_output.decoder_loss

            loss = encoder_loss + decoder_loss

            epoch_encoder_loss += encoder_loss.item()
            epoch_decoder_loss += decoder_loss.item()
            epoch_loss += loss.item()

            if epoch_loss != epoch_loss:
                raise ArithmeticError("NaN detected in train loss")

            self.callback_handler.on_train_step_end(
                training_config=self.training_config
            )

        # Allows model updates if needed
        if self.distributed:
            self.model.module.update()
        else:
            self.model.update()

        epoch_encoder_loss /= len(self.train_loader)
        epoch_decoder_loss /= len(self.train_loader)
        epoch_loss /= len(self.train_loader)

        return (epoch_loss, epoch_encoder_loss, epoch_decoder_loss)

    def save_checkpoint(self, model: BaseAE, dir_path, epoch: int):
        """Saves a checkpoint alowing to restart training from here

        Args:
            dir_path (str): The folder where the checkpoint should be saved
            epochs_signature (int): The epoch number"""

        checkpoint_dir = os.path.join(dir_path, f"checkpoint_epoch_{epoch}")

        if not os.path.exists(checkpoint_dir):
            os.makedirs(checkpoint_dir)

        # save optimizers
        torch.save(
            deepcopy(self.encoder_optimizer.state_dict()),
            os.path.join(checkpoint_dir, "encoder_optimizer.pt"),
        )
        torch.save(
            deepcopy(self.decoder_optimizer.state_dict()),
            os.path.join(checkpoint_dir, "decoder_optimizer.pt"),
        )

        # save model
        if self.distributed:
            model.module.save(checkpoint_dir)

        else:
            model.save(checkpoint_dir)

        # save training config
        self.training_config.save_json(checkpoint_dir, "training_config")



import os
from typing import Optional

import numpy as np
import torch
import torch.nn.functional as F

from ...data.datasets import BaseDataset
from ..base.base_utils import ModelOutput
from ..nn import BaseDecoder, BaseEncoder
from ..vae import VAE
from .beta_tc_vae_config import BetaTCVAEConfig


class BetaTCVAE(VAE):
    r"""
    :math:`\beta`-TCVAE model.

    Args:
        model_config (BetaTCVAEConfig): The Variational Autoencoder configuration setting the main
            parameters of the model.

        encoder (BaseEncoder): An instance of BaseEncoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        decoder (BaseDecoder): An instance of BaseDecoder (inheriting from `torch.nn.Module` which
            plays the role of decoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

    .. note::
        For high dimensional data we advice you to provide you own network architectures. With the
        provided MLP you may end up with a ``MemoryError``.
    """

    def __init__(
        self,
        model_config: BetaTCVAEConfig,
        encoder: Optional[BaseEncoder] = None,
        decoder: Optional[BaseDecoder] = None,
    ):

        VAE.__init__(self, model_config=model_config, encoder=encoder, decoder=decoder)

        self.model_name = "BetaTCVAE"
        self.alpha = model_config.alpha
        self.beta = model_config.beta
        self.gamma = model_config.gamma
        self.use_mss = model_config.use_mss

    def forward(self, inputs: BaseDataset, **kwargs):
        """
        The VAE model

        Args:
            inputs (BaseDataset): The training dataset with labels

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters

        """

        x = inputs["data"]

        dataset_size = kwargs.pop("dataset_size", x.shape[0])

        encoder_output = self.encoder(x)

        mu, log_var = encoder_output.embedding, encoder_output.log_covariance

        std = torch.exp(0.5 * log_var)
        z, _ = self._sample_gauss(mu, std)
        recon_x = self.decoder(z)["reconstruction"]

        loss, recon_loss, kld = self.loss_function(
            recon_x, x, mu, log_var, z, dataset_size
        )

        output = ModelOutput(
            recon_loss=recon_loss,
            reg_loss=kld,
            loss=loss,
            recon_x=recon_x,
            z=z,
        )

        return output

    def loss_function(self, recon_x, x, mu, log_var, z, dataset_size):

        if self.model_config.reconstruction_loss == "mse":

            recon_loss = F.mse_loss(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        elif self.model_config.reconstruction_loss == "bce":

            recon_loss = F.binary_cross_entropy(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        log_q_z_given_x = self._compute_log_gauss_density(z, mu, log_var).sum(
            dim=-1
        )  # [B]

        log_prior = self._compute_log_gauss_density(
            z, torch.zeros_like(z), torch.zeros_like(z)
        ).sum(
            dim=-1
        )  # [B]

        log_q_batch_perm = self._compute_log_gauss_density(
            z.reshape(z.shape[0], 1, -1),
            mu.reshape(1, z.shape[0], -1),
            log_var.reshape(1, z.shape[0], -1),
        )  # [B x B x Latent_dim]

        if self.use_mss:
            logiw_mat = self._log_importance_weight_matrix(z.shape[0], dataset_size).to(
                z.device
            )
            log_q_z = torch.logsumexp(
                logiw_mat + log_q_batch_perm.sum(dim=-1), dim=-1
            )  # MMS [B]
            log_prod_q_z = (
                torch.logsumexp(
                    logiw_mat.reshape(z.shape[0], z.shape[0], -1) + log_q_batch_perm,
                    dim=1,
                )
            ).sum(
                dim=-1
            )  # MMS [B]

        else:
            log_q_z = torch.logsumexp(log_q_batch_perm.sum(dim=-1), dim=-1) - torch.log(
                torch.tensor([z.shape[0] * dataset_size]).to(z.device)
            )  # MWS [B]
            log_prod_q_z = (
                torch.logsumexp(log_q_batch_perm, dim=1)
                - torch.log(torch.tensor([z.shape[0] * dataset_size]).to(z.device))
            ).sum(
                dim=-1
            )  # MWS [B]

        mutual_info_loss = log_q_z_given_x - log_q_z
        TC_loss = log_q_z - log_prod_q_z
        dimension_wise_KL = log_prod_q_z - log_prior

        return (
            (
                recon_loss
                + self.alpha * mutual_info_loss
                + self.beta * TC_loss
                + self.gamma * dimension_wise_KL
            ).mean(dim=0),
            recon_loss.mean(dim=0),
            (
                self.alpha * mutual_info_loss
                + self.beta * TC_loss
                + self.gamma * dimension_wise_KL
            ).mean(dim=0),
        )

    def _sample_gauss(self, mu, std):
        # Reparametrization trick
        # Sample N(0, I)
        eps = torch.randn_like(std)
        return mu + eps * std, eps

    def _compute_log_gauss_density(self, z, mu, log_var):
        """element-wise computation"""
        return -0.5 * (
            torch.log(torch.tensor([2 * np.pi]).to(z.device))
            + log_var
            + (z - mu) ** 2 * torch.exp(-log_var)
        )

    def _log_importance_weight_matrix(self, batch_size, dataset_size):
        """Compute importance weigth matrix for MSS
        Code from (https://github.com/rtqichen/beta-tcvae/blob/master/vae_quant.py)
        """

        N = dataset_size
        M = batch_size - 1
        strat_weight = (N - M) / (N * M)
        W = torch.Tensor(batch_size, batch_size).fill_(1 / M)
        W.view(-1)[:: M + 1] = 1 / N
        W.view(-1)[1 :: M + 1] = strat_weight
        W[M - 1, 0] = strat_weight
        return W.log()



"""A collection of Neural nets used to perform the benchmark on CIFAR"""

from .convnets import *
from .resnets import *

__all__ = [
    "Encoder_Conv_AE_CIFAR",
    "Encoder_Conv_VAE_CIFAR",
    "Encoder_Conv_SVAE_CIFAR",
    "Decoder_Conv_AE_CIFAR",
    "Discriminator_Conv_CIFAR",
    "Encoder_ResNet_AE_CIFAR",
    "Encoder_ResNet_VAE_CIFAR",
    "Encoder_ResNet_SVAE_CIFAR",
    "Encoder_ResNet_VQVAE_CIFAR",
    "Decoder_ResNet_AE_CIFAR",
    "Decoder_ResNet_VQVAE_CIFAR",
]



from pydantic.dataclasses import dataclass
from typing_extensions import Literal

from ..base import BaseSamplerConfig


@dataclass
class TwoStageVAESamplerConfig(BaseSamplerConfig):
    """Two Stage VAE sampler config class.

    Parameters:
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        second_stage_depth (int): The number of layers in the second stage VAE. Default: 2
        second_layers_dim (int): The size of the fully connected layer to used un the second VAE
            architecture.
    """

    reconstruction_loss: Literal["bce", "mse"] = "mse"
    second_stage_depth: int = 2
    second_layers_dim: int = 1024



"""This module is the implementation of the Vector Quantized VAE proposed in 
(https://arxiv.org/abs/1711.00937).

Available samplers
-------------------

Normalizing flows sampler to come.

.. autosummary::
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .vq_vae_config import VQVAEConfig
from .vq_vae_model import VQVAE

__all__ = ["VQVAE", "VQVAEConfig"]



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class FactorVAEConfig(VAEConfig):
    r"""
    FactorVAE model config config class

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        gamma (float): The balancing factor before the Total Correlation. Default: 0.5
    """
    gamma: float = 2.0
    uses_default_discriminator: bool = True
    discriminator_input_dim: int = None



import torch

from ...models import VAMP
from ..base import BaseSampler
from .vamp_sampler_config import VAMPSamplerConfig


class VAMPSampler(BaseSampler):
    """Sampling from the VAMP prior.

    Args:
        model (VAMP): The vae model to sample from.
        sampler_config (BaseSamplerConfig): An instance of BaseSamplerConfig in which any sampler's
            parameters is made available. If None a default configuration is used. Default: None

    """

    def __init__(self, model: VAMP, sampler_config: VAMPSamplerConfig = None):

        assert isinstance(model, VAMP), "This sampler is only suitable for VAMP model"

        if sampler_config is None:
            sampler_config = VAMPSamplerConfig()

        BaseSampler.__init__(self, model=model, sampler_config=sampler_config)

    def sample(
        self,
        num_samples: int = 1,
        batch_size: int = 500,
        output_dir: str = None,
        return_gen: bool = True,
        save_sampler_config: bool = False,
    ) -> torch.Tensor:
        """Main sampling function of the sampler.

        Args:
            num_samples (int): The number of samples to generate
            batch_size (int): The batch size to use during sampling
            output_dir (str): The directory where the images will be saved. If does not exist the
                folder is created. If None: the images are not saved. Defaults: None.
            return_gen (bool): Whether the sampler should directly return a tensor of generated
                data. Default: True.
            save_sampler_config (bool): Whether to save the sampler config. It is saved in
                output_dir

        Returns:
            ~torch.Tensor: The generated images
        """

        batch_size = min(self.model.idle_input.shape[0], batch_size)

        full_batch_nbr = int(num_samples / batch_size)
        last_batch_samples_nbr = num_samples % batch_size

        x_gen_list = []

        for i in range(full_batch_nbr):
            means = self.model.pseudo_inputs(self.model.idle_input.to(self.device))[
                :batch_size
            ].reshape((batch_size,) + self.model.model_config.input_dim)

            encoder_output = self.model.encoder(means)
            mu, log_var = (
                encoder_output.embedding,
                torch.tanh(encoder_output.log_covariance),
            )
            std = torch.exp(0.5 * log_var)
            eps = torch.randn_like(std)
            z = mu + eps * std

            x_gen = self.model.decoder(z)["reconstruction"].detach()

            if output_dir is not None:
                for j in range(batch_size):
                    self.save_img(
                        x_gen[j], output_dir, "%08d.png" % int(batch_size * i + j)
                    )

            x_gen_list.append(x_gen)

        if last_batch_samples_nbr > 0:
            means = self.model.pseudo_inputs(self.model.idle_input.to(self.device))[
                :last_batch_samples_nbr
            ].reshape((last_batch_samples_nbr,) + self.model.model_config.input_dim)

            encoder_output = self.model.encoder(means)
            mu, log_var = (
                encoder_output.embedding,
                torch.tanh(encoder_output.log_covariance),
            )
            std = torch.exp(0.5 * log_var)
            eps = torch.randn_like(std)
            z = mu + eps * std

            x_gen = self.model.decoder(z)["reconstruction"].detach()

            if output_dir is not None:
                for j in range(last_batch_samples_nbr):
                    self.save_img(
                        x_gen[j],
                        output_dir,
                        "%08d.png" % int(batch_size * full_batch_nbr + j),
                    )

            x_gen_list.append(x_gen)

        if save_sampler_config:
            self.save(output_dir)

        if return_gen:
            return torch.cat(x_gen_list, dim=0)



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class CIWAEConfig(VAEConfig):
    """Combination IWAE model config class.

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        number_samples (int): Number of samples to use on the Monte-Carlo estimation
        beta (float): The value of the factor in the convex combination of the VAE and IWAE ELBO.
            Default: 0.5.
    """

    number_samples: int = 10
    beta: float = 0.5

    def __post_init_post_parse__(self):
        assert 0 <= self.beta <= 1, f"Beta parameter must be in [0-1]. Got {self.beta}."



import logging
from typing import Optional, Union

import numpy as np
import torch

from ..models import BaseAE
from ..samplers import *
from ..trainers import BaseTrainerConfig
from .base_pipeline import Pipeline

logger = logging.getLogger(__name__)

# make it print to the console.
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class GenerationPipeline(Pipeline):
    """
    This Pipeline provides an end to end way to generate samples from a trained VAE model. It only
    needs a :class:`XGen.models` to sample from and a smapler configuration.

    Parameters:

        model (Optional[BaseAE]): An instance of :class:`~XGen.models.BaseAE` you want to train.
            If None, a default :class:`~XGen.models.VAE` model is used. Default: None.

        training_config (Optional[BaseTrainerConfig]): An instance of
            :class:`~XGen.trainers.BaseTrainerConfig` stating the training
            parameters. If None, a default configuration is used.
    """

    def __init__(
        self,
        model: Optional[BaseAE],
        sampler_config: Optional[BaseSamplerConfig] = None,
    ):

        if sampler_config is None:
            sampler_config = NormalSamplerConfig()

        if sampler_config.name == "NormalSamplerConfig":
            sampler = NormalSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "GaussianMixtureSamplerConfig":
            sampler = GaussianMixtureSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "IAFSamplerConfig":
            sampler = IAFSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "MAFSamplerConfig":
            sampler = MAFSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "RHVAESamplerConfig":
            sampler = RHVAESampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "PixelCNNSamplerConfig":
            sampler = PixelCNNSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "TwoStageVAESamplerConfig":
            sampler = TwoStageVAESampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "VAMPSamplerConfig":
            sampler = VAMPSampler(model=model, sampler_config=sampler_config)

        elif sampler_config.name == "HypersphereUniformSamplerConfig":
            sampler = HypersphereUniformSampler(
                model=model, sampler_config=sampler_config
            )

        elif sampler_config.name == "PoincareDiskSamplerConfig":
            sampler = PoincareDiskSampler(model=model, sampler_config=sampler_config)

        else:
            raise NotImplementedError(
                "Unrecognized sampler config name... Check that that the sampler_config name "
                f"is written correctly. Got '{sampler_config.name}'."
            )

        self.sampler = sampler

    def __call__(
        self,
        num_samples: int = 1,
        batch_size: int = 500,
        output_dir: str = None,
        return_gen: bool = True,
        save_sampler_config: bool = False,
        train_data: Union[np.ndarray, torch.Tensor] = None,
        eval_data: Union[np.ndarray, torch.Tensor] = None,
        training_config: BaseTrainerConfig = None,
    ):
        """
        Launch the model training on the provided data.

        Args:
            num_samples (int): The number of samples to generate

            batch_size (int): The batch size to use during sampling

            output_dir (str): The directory where the images will be saved. If does not exist the
                folder is created. If None: the images are not saved. Defaults: None.

            return_gen (bool): Whether the sampler should directly return a tensor of generated
                data. Default: True.

            training_data (Union[~numpy.ndarray, ~torch.Tensor]): The training data as a
                :class:`numpy.ndarray` or :class:`torch.Tensor` of shape (mini_batch x
                n_channels x ...) if the sampler needs to be trained (e.g. flow based samplers).
                Default: None.

            eval_data (Optional[Union[~numpy.ndarray, ~torch.Tensor]]): The evaluation data as a
                :class:`numpy.ndarray` or :class:`torch.Tensor` of shape (mini_batch x
                n_channels x ...) if the sampler needs to be trained (e.g. flow based samplers).
                Default: None.

            training_config (BaseTrainerConfig): the training config to use if the sampler needs to
                be trained (e.g. flow based samplers). Default: None.

            callbacks (List[~XGen.trainers.training_callbacks.TrainingCallbacks]):
                A list of callbacks to use during training.
        """

        # Fit the sampler
        self.sampler.fit(
            train_data=train_data, eval_data=eval_data, training_config=training_config
        )

        # Generate data
        generated_samples = self.sampler.sample(
            num_samples=num_samples,
            batch_size=batch_size,
            output_dir=output_dir,
            return_gen=return_gen,
            save_sampler_config=save_sampler_config,
        )

        return generated_samples



"""
Implementation of the Inverse Autoregressive Flows (IAF) proposed
in (https://arxiv.org/abs/1606.04934).
"""

from .iaf_config import IAFConfig
from .iaf_model import IAF

__all__ = ["IAF", "IAFConfig"]



"""This module is the implementation of a Poincaré Disk Variational Autoencoder
(https://arxiv.org/abs/1901.06033).

Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.PoincareDiskSampler
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .pvae_config import PoincareVAEConfig
from .pvae_model import PoincareVAE

__all__ = ["PoincareVAE", "PoincareVAEConfig"]



"""This module is the implementation of the Hamiltonian VAE proposed in
(https://arxiv.org/abs/1805.11328).
This model combines Hamiltonian Monte Carlo smapling and normalizing flows together to improve the 
true posterior estimate within the VAE framework.

Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .hvae_config import HVAEConfig
from .hvae_model import HVAE

__all__ = ["HVAE", "HVAEConfig"]



import importlib
import io
import logging
from collections import OrderedDict
from typing import Any, Tuple

try:
    import pickle5 as pickle
except:
    import pickle
import torch

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)

model_card_template = """---
language: en
tags:
- XGen
license: apache-2.0
---

### Downloading this model from the Hub
This model was trained with XGen. It can be downloaded or reloaded using the method `load_from_hf_hub`
```python
>>> from XGen.models import AutoModel
>>> model = AutoModel.load_from_hf_hub(hf_hub_path="your_hf_username/repo_name")
```
"""


def hf_hub_is_available():
    return importlib.util.find_spec("huggingface_hub") is not None


class ModelOutput(OrderedDict):
    """Base ModelOutput class fixing the output type from the models. This class is inspired from
    the ``ModelOutput`` class from hugginface transformers library"""

    def __getitem__(self, k):
        if isinstance(k, str):
            self_dict = {k: v for (k, v) in self.items()}
            return self_dict[k]
        else:
            return self.to_tuple()[k]

    def __setattr__(self, name, value):
        super().__setitem__(name, value)
        super().__setattr__(name, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().__setattr__(key, value)

    def to_tuple(self) -> Tuple[Any]:
        """
        Convert self to a tuple containing all the attributes/keys that are not ``None``.
        """
        return tuple(self[k] for k in self.keys())


class CPU_Unpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if module == "torch.storage" and name == "_load_from_bytes":
            return lambda b: torch.load(io.BytesIO(b), map_location="cpu")
        else:
            return super().find_class(module, name)



"""
Implementation of the Masked Autoregressive Flow (MAF) proposed
in (https://arxiv.org/abs/1502.03509)
"""

from .maf_config import MAFConfig
from .maf_model import MAF

__all__ = ["MAF", "MAFConfig"]



import os

import numpy as np
import torch
import torch.nn as nn

from XGen.models.base.base_utils import ModelOutput

from ..base import BaseNF
from ..layers import MaskedLinear
from .made_config import MADEConfig


class MADE(BaseNF):
    """Masked Autoencoder model

    Args:
        model_config (MADEConfig): The MADE model configuration setting the main parameters of the
            model
    """

    def __init__(self, model_config: MADEConfig):

        BaseNF.__init__(self, model_config=model_config)

        self.net = []
        self.m = {}
        self.model_config = model_config
        self.input_dim = np.prod(model_config.input_dim)
        self.output_dim = np.prod(model_config.output_dim)
        self.hidden_sizes = model_config.hidden_sizes
        self.model_name = "MADE"

        if model_config.input_dim is None:
            raise AttributeError(
                "No input dimension provided !"
                "'input_dim' parameter of MADEConfig instance must be set to 'data_shape' "
                "where the shape of the data is (C, H, W ..)]. Unable to build network"
                "automatically"
            )

        if model_config.output_dim is None:
            raise AttributeError(
                "No input dimension provided !"
                "'output_dim' parameter of MADEConfig instance must be set to 'data_shape' "
                "where the shape of the data is (C, H, W ..)]. Unable to build network"
                "automatically"
            )

        hidden_sizes = [self.input_dim] + model_config.hidden_sizes + [self.output_dim]

        masks = self._make_mask(ordering=self.model_config.degrees_ordering)

        for inp, out, mask in zip(hidden_sizes[:-1], hidden_sizes[1:-1], masks[:-1]):

            self.net.extend([MaskedLinear(inp, out, mask), nn.ReLU()])

        # outputs mean and logvar
        self.net.extend(
            [
                MaskedLinear(
                    self.hidden_sizes[-1], 2 * self.output_dim, masks[-1].repeat(2, 1)
                )
            ]
        )

        self.net = nn.Sequential(*self.net)

    def _make_mask(self, ordering="sequential"):

        # Get degrees for mask creation

        if ordering == "sequential":
            self.m[-1] = torch.arange(self.input_dim)
            for i in range(len(self.hidden_sizes)):
                self.m[i] = torch.arange(self.hidden_sizes[i]) % (self.input_dim - 1)

        else:
            self.m[-1] = torch.randperm(self.input_dim)
            for i in range(len(self.hidden_sizes)):
                self.m[i] = torch.randint(
                    self.m[-1].min(), self.input_dim - 1, (self.hidden_sizes[i],)
                )

        masks = []
        for i in range(len(self.hidden_sizes)):
            masks += [(self.m[i].unsqueeze(-1) >= self.m[i - 1].unsqueeze(0)).float()]

        masks.append(
            (
                self.m[len(self.hidden_sizes) - 1].unsqueeze(0)
                < self.m[-1].unsqueeze(-1)
            ).float()
        )

        return masks

    def forward(self, x: torch.tensor, **kwargs) -> ModelOutput:
        """The input data is transformed toward the prior

        Args:
            inputs (torch.Tensor): An input tensor

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters
        """
        net_output = self.net(x.reshape(x.shape[0], -1))

        mu = net_output[:, : self.input_dim]
        log_var = net_output[:, self.input_dim :]

        return ModelOutput(mu=mu, log_var=log_var)



import os
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ...data.datasets import BaseDataset
from ..base.base_utils import ModelOutput
from ..nn import BaseDecoder, BaseEncoder
from ..normalizing_flows import (
    PlanarFlow,
    PlanarFlowConfig,
    RadialFlow,
    RadialFlowConfig,
)
from ..vae import VAE
from .vae_lin_nf_config import VAE_LinNF_Config


class VAE_LinNF(VAE):
    """Variational Auto Encoder with linear Normalizing Flows model.

    Args:
        model_config(VAE_LinNF_Config): The Variational Autoencoder configuration seting the main
        parameters of the model

        encoder (BaseEncoder): An instance of BaseEncoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        decoder (BaseDecoder): An instance of BaseDecoder (inheriting from `torch.nn.Module` which
            plays the role of decoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        flows (List[str]): A list of strings corresponding to the class of each flow to be applied.
            Default: Empty list (no flow is applied).

    .. note::
        For high dimensional data we advice you to provide you own network architectures. With the
        provided MLP you may end up with a ``MemoryError``.
    """

    def __init__(
        self,
        model_config: VAE_LinNF_Config,
        encoder: Optional[BaseEncoder] = None,
        decoder: Optional[BaseDecoder] = None,
    ):

        VAE.__init__(self, model_config=model_config, encoder=encoder, decoder=decoder)

        self.model_name = "VAE_LinNF"

        self.net = []
        for flow in model_config.flows:
            if flow == "Planar":
                flow_config = PlanarFlowConfig(
                    input_dim=(model_config.latent_dim,), activation="tanh"
                )
                self.net.append(PlanarFlow(flow_config))

            elif flow == "Radial":
                flow_config = RadialFlowConfig(input_dim=(model_config.latent_dim,))
                self.net.append(RadialFlow(flow_config))

        self.net = nn.ModuleList(self.net)

    def forward(self, inputs: BaseDataset, **kwargs):
        """
        The VAE with linear normalizing flows model

        Args:
            inputs (BaseDataset): The training dataset with labels

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters

        """

        x = inputs["data"]

        encoder_output = self.encoder(x)

        mu, log_var = encoder_output.embedding, encoder_output.log_covariance

        std = torch.exp(0.5 * log_var)
        z, _ = self._sample_gauss(mu, std)

        z0 = z

        log_abs_det_jac = torch.zeros((z0.shape[0],)).to(z.device)

        for layer in self.net:
            layer_output = layer(z)
            z = layer_output.out
            log_abs_det_jac += layer_output.log_abs_det_jac

        recon_x = self.decoder(z)["reconstruction"]

        loss, recon_loss, kld = self.loss_function(
            recon_x, x, mu, log_var, z0, z, log_abs_det_jac
        )

        output = ModelOutput(
            recon_loss=recon_loss,
            reg_loss=kld,
            loss=loss,
            recon_x=recon_x,
            z=z,
        )

        return output

    def loss_function(self, recon_x, x, mu, log_var, z0, zk, log_abs_det_jac):

        if self.model_config.reconstruction_loss == "mse":

            recon_loss = F.mse_loss(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        elif self.model_config.reconstruction_loss == "bce":

            recon_loss = F.binary_cross_entropy(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        # starting gaussian log-density
        log_prob_z0 = (
            -0.5 * (log_var + torch.pow(z0 - mu, 2) / torch.exp(log_var))
        ).sum(dim=1)

        # prior log-density
        log_prob_zk = (-0.5 * torch.pow(zk, 2)).sum(dim=1)

        KLD = log_prob_z0 - log_prob_zk - log_abs_det_jac

        return (recon_loss + KLD).mean(dim=0), recon_loss.mean(dim=0), KLD.mean(dim=0)

    def _sample_gauss(self, mu, std):
        # Reparametrization trick
        # Sample N(0, I)
        eps = torch.randn_like(std)
        return mu + eps * std, eps

    def get_nll(self, data, n_samples=1, batch_size=100):
        """
        Function computed the estimate negative log-likelihood of the model. It uses importance
        sampling method with the approximate posterior distribution. This may take a while.

        Args:
            data (torch.Tensor): The input data from which the log-likelihood should be estimated.
                Data must be of shape [Batch x n_channels x ...]
            n_samples (int): The number of importance samples to use for estimation
            batch_size (int): The batchsize to use to avoid memory issues
        """
        normal = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.model_config.latent_dim).to(data.device),
            covariance_matrix=torch.eye(self.model_config.latent_dim).to(data.device),
        )

        if n_samples <= batch_size:
            n_full_batch = 1
        else:
            n_full_batch = n_samples // batch_size
            n_samples = batch_size

        log_p = []

        for i in range(len(data)):
            x = data[i].unsqueeze(0)

            log_p_x = []

            for j in range(n_full_batch):

                x_rep = torch.cat(batch_size * [x])

                encoder_output = self.encoder(x_rep)
                mu, log_var = encoder_output.embedding, encoder_output.log_covariance

                std = torch.exp(0.5 * log_var)
                z, eps = self._sample_gauss(mu, std)

                z0 = z

                log_abs_det_jac = torch.zeros((z0.shape[0],)).to(z.device)

                for layer in self.net:
                    layer_output = layer(z)
                    z = layer_output.out
                    log_abs_det_jac += layer_output.log_abs_det_jac

                log_q_z_given_x = (
                    -0.5 * (log_var + torch.pow(z0 - mu, 2) / torch.exp(log_var))
                ).sum(dim=1) - log_abs_det_jac
                log_p_z = -0.5 * (z ** 2).sum(dim=-1)

                recon_x = self.decoder(z)["reconstruction"]

                if self.model_config.reconstruction_loss == "mse":

                    log_p_x_given_z = -0.5 * F.mse_loss(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1) - torch.tensor(
                        [np.prod(self.input_dim) / 2 * np.log(np.pi * 2)]
                    ).to(
                        data.device
                    )  # decoding distribution is assumed unit variance  N(mu, I)

                elif self.model_config.reconstruction_loss == "bce":

                    log_p_x_given_z = -F.binary_cross_entropy(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1)

                log_p_x.append(
                    log_p_x_given_z + log_p_z - log_q_z_given_x
                )  # log(2*pi) simplifies

            log_p_x = torch.cat(log_p_x)

            log_p.append((torch.logsumexp(log_p_x, 0) - np.log(len(log_p_x))).item())

        return np.mean(log_p)



from pydantic.dataclasses import dataclass
from typing_extensions import Literal

from ..base.base_config import BaseAEConfig


@dataclass
class VAEConfig(BaseAEConfig):
    """VAE config class.

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
    """

    reconstruction_loss: Literal["bce", "mse"] = "mse"



from pydantic.dataclasses import dataclass
from typing_extensions import Literal

from ..vae import VAEConfig


@dataclass
class PoincareVAEConfig(VAEConfig):
    """Poincaré VAE config class.

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        prior_distribution (str): The distribution to use as prior
            ["wrapped_normal", "riemannian_normal"]. Default: "wrapped_normal"
        posterior_distribution (str): The distribution to use as posterior
            ["wrapped_normal", "riemannian_normal"]. Default: "wrapped_normal"
        curvature (int): The curvature of the manifold. Default: 1
    """

    prior_distribution: Literal[
        "wrapped_normal", "riemannian_normal"
    ] = "wrapped_normal"
    posterior_distribution: Literal[
        "wrapped_normal", "riemannian_normal"
    ] = "wrapped_normal"
    curvature: float = 1



"""A collection of Neural nets used to perform the benchmark on MNIST"""

from .convnets import *
from .resnets import *

__all__ = [
    "Encoder_Conv_AE_MNIST",
    "Encoder_Conv_VAE_MNIST",
    "Encoder_Conv_SVAE_MNIST",
    "Decoder_Conv_AE_MNIST",
    "Discriminator_Conv_MNIST",
    "Encoder_ResNet_AE_MNIST",
    "Encoder_ResNet_VAE_MNIST",
    "Encoder_ResNet_SVAE_MNIST",
    "Encoder_ResNet_VQVAE_MNIST",
    "Decoder_ResNet_AE_MNIST",
    "Decoder_ResNet_VQVAE_MNIST",
]



import torch


def create_metric(model):
    def G(z):
        return torch.inverse(
            (
                model.M_tens.unsqueeze(0).to(z.device)
                * torch.exp(
                    -torch.norm(
                        model.centroids_tens.unsqueeze(0).to(z.device) - z.unsqueeze(1),
                        dim=-1,
                    )
                    ** 2
                    / (model.temperature ** 2)
                )
                .unsqueeze(-1)
                .unsqueeze(-1)
            ).sum(dim=1)
            + model.lbd * torch.eye(model.latent_dim).to(z.device)
        )

    return G


def create_inverse_metric(model):
    def G_inv(z):
        return (
            model.M_tens.unsqueeze(0).to(z.device)
            * torch.exp(
                -torch.norm(
                    model.centroids_tens.unsqueeze(0).to(z.device) - z.unsqueeze(1),
                    dim=-1,
                )
                ** 2
                / (model.temperature ** 2)
            )
            .unsqueeze(-1)
            .unsqueeze(-1)
        ).sum(dim=1) + model.lbd * torch.eye(model.latent_dim).to(z.device)

    return G_inv



import torch
import torch.nn as nn


class ResBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        nn.Module.__init__(self)

        self.conv_block = nn.Sequential(
            nn.ReLU(),
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(out_channels, in_channels, kernel_size=1, stride=1, padding=0),
        )

    def forward(self, x: torch.tensor) -> torch.Tensor:
        return x + self.conv_block(x)



from pydantic.dataclasses import dataclass

from ..ae import AEConfig


@dataclass
class RAE_L2_Config(AEConfig):
    """RAE_L2 config class.

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        embedding_weight (float): The factor before the L2 regularization term in the loss.
            Default: 1e-4
        reg_weight (float): The weight decay to apply.
    """

    embedding_weight: float = 1e-4
    reg_weight: float = 1e-7



"""Basic sampler sampling from a uniform distribution on the hypersphere 
in the Autoencoder's latent space.

Available models:
------------------

.. autosummary::
    ~XGen.models.SVAE
    :nosignatures:
"""

from .hypersphere_uniform_config import HypersphereUniformSamplerConfig
from .hypersphere_uniform_sampler import HypersphereUniformSampler

__all__ = ["HypersphereUniformSampler", "HypersphereUniformSamplerConfig"]



"""This module is the implementation of the BetaVAE proposed in
(https://openreview.net/pdf?id=Sy2fzU9gl).
This model adds a new parameter to the VAE loss function balancing the weight of the 
reconstruction term and KL term.


Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .beta_vae_config import BetaVAEConfig
from .beta_vae_model import BetaVAE

__all__ = ["BetaVAE", "BetaVAEConfig"]



"""This module is the implementation of the MSSSIM_VAE proposed in
(https://arxiv.org/abs/1511.06409).
This models uses a perceptual similarity metric for reconstruction.


Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .msssim_vae_config import MSSSIM_VAEConfig
from .msssim_vae_model import MSSSIM_VAE

__all__ = ["MSSSIM_VAE", "MSSSIM_VAEConfig"]



r"""This module is the implementation of the Disentangled Beta VAE proposed in
(https://arxiv.org/abs/1804.03599).
This model adds a new parameter to the :math:`\beta`-VAE loss function corresponding to the target 
value for the KL between the prior and the posterior distribution. It is progressively increased
throughout training.


Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .disentangled_beta_vae_config import DisentangledBetaVAEConfig
from .disentangled_beta_vae_model import DisentangledBetaVAE

__all__ = ["DisentangledBetaVAE", "DisentangledBetaVAEConfig"]



import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F

from ..base.base_utils import ModelOutput
from .vq_vae_config import VQVAEConfig

"""Code inspired from https://github.com/deepmind/sonnet/blob/v2/sonnet/src/nets/vqvae.py"""


class Quantizer(nn.Module):
    def __init__(self, model_config: VQVAEConfig):

        nn.Module.__init__(self)

        self.model_config = model_config

        self.embedding_dim = model_config.embedding_dim
        self.num_embeddings = model_config.num_embeddings
        self.commitment_loss_factor = model_config.commitment_loss_factor
        self.quantization_loss_factor = model_config.quantization_loss_factor

        self.embeddings = nn.Embedding(self.num_embeddings, self.embedding_dim)

        self.embeddings.weight.data.uniform_(
            -1 / self.num_embeddings, 1 / self.num_embeddings
        )

    def forward(self, z: torch.Tensor, uses_ddp: bool = False):

        distances = (
            (z.reshape(-1, self.embedding_dim) ** 2).sum(dim=-1, keepdim=True)
            + (self.embeddings.weight ** 2).sum(dim=-1)
            - 2 * z.reshape(-1, self.embedding_dim) @ self.embeddings.weight.T
        )

        closest = distances.argmin(-1).unsqueeze(-1)

        quantized_indices = closest.reshape(z.shape[0], z.shape[1], z.shape[2])

        one_hot_encoding = (
            F.one_hot(closest, num_classes=self.num_embeddings)
            .type(torch.float)
            .squeeze(1)
        )

        # quantization
        quantized = one_hot_encoding @ self.embeddings.weight
        quantized = quantized.reshape_as(z)

        commitment_loss = F.mse_loss(
            quantized.detach().reshape(-1, self.embedding_dim),
            z.reshape(-1, self.embedding_dim),
            reduction="mean",
        )

        embedding_loss = F.mse_loss(
            quantized.reshape(-1, self.embedding_dim),
            z.detach().reshape(-1, self.embedding_dim),
            reduction="mean",
        ).mean(dim=-1)

        quantized = z + (quantized - z).detach()

        loss = (
            commitment_loss * self.commitment_loss_factor
            + embedding_loss * self.quantization_loss_factor
        )
        quantized = quantized.permute(0, 3, 1, 2)

        output = ModelOutput(
            quantized_vector=quantized,
            quantized_indices=quantized_indices.unsqueeze(1),
            loss=loss,
        )

        return output


class QuantizerEMA(nn.Module):
    def __init__(self, model_config: VQVAEConfig):

        nn.Module.__init__(self)

        self.model_config = model_config

        self.embedding_dim = model_config.embedding_dim
        self.num_embeddings = model_config.num_embeddings
        self.commitment_loss_factor = model_config.commitment_loss_factor
        self.decay = model_config.decay

        embeddings = torch.empty(self.num_embeddings, self.embedding_dim)

        embeddings.data.uniform_(-1 / self.num_embeddings, 1 / self.num_embeddings)

        self.register_buffer("cluster_size", torch.zeros(self.num_embeddings))
        self.register_buffer(
            "ema_embed", torch.zeros(self.num_embeddings, self.embedding_dim)
        )
        self.ema_embed.data.uniform_(-1 / self.num_embeddings, 1 / self.num_embeddings)

        self.register_buffer("embeddings", embeddings)

    def forward(self, z: torch.Tensor, uses_ddp: bool = False):

        distances = (
            (z.reshape(-1, self.embedding_dim) ** 2).sum(dim=-1, keepdim=True)
            + (self.embeddings ** 2).sum(dim=-1)
            - 2 * z.reshape(-1, self.embedding_dim) @ self.embeddings.T
        )

        closest = distances.argmin(-1).unsqueeze(-1)

        quantized_indices = closest.reshape(z.shape[0], z.shape[1], z.shape[2])

        one_hot_encoding = (
            F.one_hot(closest, num_classes=self.num_embeddings)
            .type(torch.float)
            .squeeze(1)
        )

        # quantization
        quantized = one_hot_encoding @ self.embeddings
        quantized = quantized.reshape_as(z)

        if self.training:

            n_i = torch.sum(one_hot_encoding, dim=0)

            if uses_ddp:
                dist.all_reduce(n_i)

            self.cluster_size = self.cluster_size * self.decay + n_i * (1 - self.decay)

            dw = one_hot_encoding.T @ z.reshape(-1, self.embedding_dim)

            if uses_ddp:
                dist.all_reduce(dw)

            ema_embed = self.ema_embed * self.decay + dw * (1 - self.decay)

            n = torch.sum(self.cluster_size)

            self.cluster_size = (
                (self.cluster_size + 1e-5) / (n + self.num_embeddings * 1e-5) * n
            )

            self.embeddings.data.copy_(ema_embed / self.cluster_size.unsqueeze(-1))
            self.ema_embed.data.copy_(ema_embed)

        commitment_loss = F.mse_loss(
            quantized.detach().reshape(-1, self.embedding_dim),
            z.reshape(-1, self.embedding_dim),
            reduction="mean",
        )

        quantized = z + (quantized - z).detach()

        loss = commitment_loss * self.commitment_loss_factor
        quantized = quantized.permute(0, 3, 1, 2)

        output = ModelOutput(
            quantized_vector=quantized,
            quantized_indices=quantized_indices.unsqueeze(1),
            loss=loss,
        )

        return output



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class VAE_IAF_Config(VAEConfig):
    """VAE with Inverse Autoregressive Normalizing Flow config class.

    Parameters:
        input_dim (int): The input_data dimension
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'.
        n_made_blocks (int): The number of :class:`~XGen.models.normalizing_flows.MADE` models
            to consider in the IAF used in the VAE. Default: 2.
        n_hidden_in_made (int): The number of hidden layers in the
            :class:`~XGen.models.normalizing_flows.MADE` models composing the IAF in the VAE.
            Default: 3.
        hidden_size (list): The number of unit in each hidder layerin the
            :class:`~XGen.models.normalizing_flows.MADE` model. The same number of
            units is used across the `n_hidden_in_made` and `n_made_blocks`. Default: 128.
    """

    n_made_blocks: int = 2
    n_hidden_in_made: int = 3
    hidden_size: int = 128



""" Here are implemented the trainers used to train the Autoencoder models
"""

from .adversarial_trainer import AdversarialTrainer, AdversarialTrainerConfig
from .base_trainer import BaseTrainer, BaseTrainerConfig
from .coupled_optimizer_adversarial_trainer import (
    CoupledOptimizerAdversarialTrainer,
    CoupledOptimizerAdversarialTrainerConfig,
)
from .coupled_optimizer_trainer import (
    CoupledOptimizerTrainer,
    CoupledOptimizerTrainerConfig,
)

__all__ = [
    "BaseTrainer",
    "BaseTrainerConfig",
    "CoupledOptimizerTrainer",
    "CoupledOptimizerTrainerConfig",
    "AdversarialTrainer",
    "AdversarialTrainerConfig",
    "CoupledOptimizerAdversarialTrainer",
    "CoupledOptimizerAdversarialTrainerConfig",
]



"""This module is the implementation of a Variational Auto Encoder with Normalizing Flow
(https://arxiv.org/pdf/1505.05770.pdf).

Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:
"""

from .vae_lin_nf_config import VAE_LinNF_Config
from .vae_lin_nf_model import VAE_LinNF

__all__ = ["VAE_LinNF", "VAE_LinNF_Config"]



from pydantic.dataclasses import dataclass

from ..base import BaseNFConfig


@dataclass
class PlanarFlowConfig(BaseNFConfig):
    """This is the PlanarFlow model configuration instance.

    Parameters:
        input_dim (tuple): The input data dimension. Default: None.
        activation (str): The activation function to be applied. Choices: ['linear', 'tanh', 'elu'].
            Default: 'tanh'.
    """

    activation: str = "tanh"

    def __post_init_post_parse__(self):
        assert self.activation in ["linear", "tanh", "elu"], (
            f"'{self.activation}' doesn't correspond to an activation handled by the model. "
            "Available activations ['linear', 'tanh', 'elu']"
        )



from typing import Union

import torch.nn as nn
from pydantic.dataclasses import dataclass

from ..base_trainer import BaseTrainerConfig


@dataclass
class CoupledOptimizerTrainerConfig(BaseTrainerConfig):
    """
    CoupledOptimizerTrainer config class.

    Parameters:

        output_dir (str): The directory where model checkpoints, configs and final
            model will be stored. Default: None.
        per_device_train_batch_size (int): The number of training samples per batch and per device.
            Default 64
        per_device_eval_batch_size (int): The number of evaluation samples per batch and per device.
            Default 64
        num_epochs (int): The maximal number of epochs for training. Default: 100
        train_dataloader_num_workers (int): Number of subprocesses to use for train data loading.
            0 means that the data will be loaded in the main process. Default: 0
        eval_dataloader_num_workers (int): Number of subprocesses to use for evaluation data
            loading. 0 means that the data will be loaded in the main process. Default: 0
        encoder_optimizer_cls (str): The name of the `torch.optim.Optimizer` used for
            the training of the encoder. Default: :class:`~torch.optim.Adam`.
        encoder_optimizer_params (dict): A dict containing the parameters to use for the
            `torch.optim.Optimizer` for the encoder. If None, uses the default parameters.
            Default: None.
        encoder_scheduler_cls (str): The name of the `torch.optim.lr_scheduler` used for
            the training of the encoder. Default :class:`~torch.optim.Adam`.
        encoder_scheduler_params (dict): A dict containing the parameters to use for the
            `torch.optim.le_scheduler`  for the encoder. If None, uses the default parameters.
            Default: None.
        decoder_optimizer_cls (str): The name of the `torch.optim.Optimizer` used for
            the training of the decoder. Default: :class:`~torch.optim.Adam`.
        decoder_optimizer_params (dict): A dict containing the parameters to use for the
            `torch.optim.Optimizer` for the decoder. If None, uses the default parameters.
            Default: None.
        decoder_scheduler_cls (str): The name of the `torch.optim.lr_scheduler` used for
            the training of the decoder. Default :class:`~torch.optim.Adam`.
        decoder_scheduler_params (dict): A dict containing the parameters to use for the
            `torch.optim.le_scheduler`  for the decoder. If None, uses the default parameters.
            Default: None.
        encoder_learning_rate (int): The learning rate applied to the `Optimizer` for the encoder.
            Default: 1e-4
        decoder_learning_rate (int): The learning rate applied to the `Optimizer` for the
            decoder. Default: 1e-4
        steps_saving (int): A model checkpoint will be saved every `steps_saving` epoch.
            Default: None
        steps_saving (int): A prediction using the best model will be run every `steps_predict`
            epoch. Default: None
        keep_best_on_train (bool): Whether to keep the best model on the train set. Default: False
        seed (int): The random seed for reproducibility
        no_cuda (bool): Disable `cuda` training. Default: False
        world_size (int): The total number of process to run. Default: -1
        local_rank (int): The rank of the node for distributed training. Default: -1
        rank (int): The rank of the process for distributed training. Default: -1
        dist_backend (str): The distributed backend to use. Default: 'nccl'
        master_addr (str): The master address for distributed training. Default: 'localhost'
        master_port (str): The master port for distributed training. Default: '12345'

    """

    encoder_optimizer_cls: str = "Adam"
    encoder_optimizer_params: Union[dict, None] = None
    encoder_scheduler_cls: str = None
    encoder_scheduler_params: Union[dict, None] = None
    decoder_optimizer_cls: str = "Adam"
    decoder_optimizer_params: Union[dict, None] = None
    decoder_scheduler_cls: str = None
    decoder_scheduler_params: Union[dict, None] = None
    encoder_learning_rate: float = 1e-4
    decoder_learning_rate: float = 1e-4

    def __post_init_post_parse__(self):
        """Check compatibilty"""

        # encoder optimizer and scheduler
        try:
            import torch.optim as optim

            encoder_optimizer_cls = getattr(optim, self.encoder_optimizer_cls)
        except AttributeError as e:
            raise AttributeError(
                f"Unable to import `{self.encoder_optimizer_cls}` encoder optimizer "
                "from 'torch.optim'. Check spelling and that it is part of "
                "'torch.optim.Optimizers.'"
            )
        if self.encoder_optimizer_params is not None:
            try:
                encoder_optimizer = encoder_optimizer_cls(
                    nn.Linear(2, 2).parameters(),
                    lr=self.encoder_learning_rate,
                    **self.encoder_optimizer_params,
                )
            except TypeError as e:
                raise TypeError(
                    "Error in optimizer's parameters. Check that the provided dict contains only "
                    f"keys and values suitable for `{encoder_optimizer_cls}` optimizer. "
                    f"Got {self.encoder_optimizer_params} as parameters.\n"
                    f"Exception raised: {type(e)} with message: " + str(e)
                ) from e
        else:
            encoder_optimizer = encoder_optimizer_cls(
                nn.Linear(2, 2).parameters(), lr=self.encoder_learning_rate
            )

        if self.encoder_scheduler_cls is not None:
            try:
                import torch.optim.lr_scheduler as schedulers

                encoder_scheduder_cls = getattr(schedulers, self.encoder_scheduler_cls)
            except AttributeError as e:
                raise AttributeError(
                    f"Unable to import `{self.encoder_scheduler_cls}` encoder scheduler from "
                    "'torch.optim.lr_scheduler'. Check spelling and that it is part of "
                    "'torch.optim.lr_scheduler.'"
                )

            if self.encoder_scheduler_params is not None:
                try:
                    encoder_scheduder_cls(
                        encoder_optimizer, **self.encoder_scheduler_params
                    )
                except TypeError as e:
                    raise TypeError(
                        "Error in scheduler's parameters. Check that the provided dict contains only "
                        f"keys and values suitable for `{encoder_scheduder_cls}` scheduler. "
                        f"Got {self.encoder_scheduler_params} as parameters.\n"
                        f"Exception raised: {type(e)} with message: " + str(e)
                    ) from e

        # decoder optimizer and scheduler
        try:
            decoder_optimizer_cls = getattr(optim, self.decoder_optimizer_cls)
        except AttributeError as e:
            raise AttributeError(
                f"Unable to import `{self.decoder_optimizer_cls}` decoder optimizer "
                "from 'torch.optim'. Check spelling and that it is part of "
                "'torch.optim.Optimizers.'"
            )
        if self.decoder_optimizer_params is not None:
            try:
                decoder_optimizer = decoder_optimizer_cls(
                    nn.Linear(2, 2).parameters(),
                    lr=self.decoder_learning_rate,
                    **self.decoder_optimizer_params,
                )
            except TypeError as e:
                raise TypeError(
                    "Error in optimizer's parameters. Check that the provided dict contains only "
                    f"keys and values suitable for `{decoder_optimizer_cls}` optimizer. "
                    f"Got {self.decoder_optimizer_params} as parameters.\n"
                    f"Exception raised: {type(e)} with message: " + str(e)
                ) from e
        else:
            decoder_optimizer = decoder_optimizer_cls(
                nn.Linear(2, 2).parameters(), lr=self.decoder_learning_rate
            )

        if self.decoder_scheduler_cls is not None:
            try:
                import torch.optim.lr_scheduler as schedulers

                decoder_scheduder_cls = getattr(schedulers, self.decoder_scheduler_cls)
            except AttributeError as e:
                raise AttributeError(
                    f"Unable to import `{self.decoder_scheduler_cls}` decoder scheduler from "
                    "'torch.optim.lr_scheduler'. Check spelling and that it is part of "
                    "'torch.optim.lr_scheduler.'"
                )

            if self.decoder_scheduler_params is not None:
                try:
                    decoder_scheduder_cls(
                        decoder_optimizer, **self.decoder_scheduler_params
                    )
                except TypeError as e:
                    raise TypeError(
                        "Error in scheduler's parameters. Check that the provided dict contains only "
                        f"keys and values suitable for `{decoder_scheduder_cls}` scheduler. "
                        f"Got {self.decoder_scheduler_params} as parameters.\n"
                        f"Exception raised: {type(e)} with message: " + str(e)
                    ) from e



import inspect
import logging
import os
import warnings
from collections import deque
from copy import deepcopy
from typing import Optional

import cloudpickle
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import grad

from ...customexception import BadInheritanceError
from ...data.datasets import BaseDataset
from ..base.base_utils import CPU_Unpickler, ModelOutput, hf_hub_is_available
from ..nn import BaseDecoder, BaseEncoder, BaseMetric
from ..nn.default_architectures import Metric_MLP
from ..vae import VAE
from .rhvae_config import RHVAEConfig
from .rhvae_utils import create_inverse_metric, create_metric

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class RHVAE(VAE):
    r"""
    Riemannian Hamiltonian VAE model.


    Args:
        model_config (RHVAEConfig): A model configuration setting the main parameters of the model.

        encoder (BaseEncoder): An instance of BaseEncoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        decoder (BaseDecoder): An instance of BaseDecoder (inheriting from `torch.nn.Module` which
            plays the role of decoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.


    .. note::
        For high dimensional data we advice you to provide you own network architectures. With the
        provided MLP you may end up with a ``MemoryError``.
    """

    def __init__(
        self,
        model_config: RHVAEConfig,
        encoder: Optional[BaseEncoder] = None,
        decoder: Optional[BaseDecoder] = None,
        metric: Optional[BaseMetric] = None,
    ):

        VAE.__init__(self, model_config=model_config, encoder=encoder, decoder=decoder)

        self.model_name = "RHVAE"

        if metric is None:
            metric = Metric_MLP(model_config)
            self.model_config.uses_default_metric = True

        else:
            self.model_config.uses_default_metric = False

        self.set_metric(metric)

        self.temperature = nn.Parameter(
            torch.Tensor([model_config.temperature]), requires_grad=False
        )
        self.lbd = nn.Parameter(
            torch.Tensor([model_config.regularization]), requires_grad=False
        )
        self.beta_zero_sqrt = nn.Parameter(
            torch.Tensor([model_config.beta_zero]), requires_grad=False
        )
        self.n_lf = model_config.n_lf
        self.eps_lf = model_config.eps_lf

        # this is used to store the matrices and centroids throughout training for
        # further use in metric update (L is the cholesky decomposition of M)
        self.M = deque(maxlen=100)
        self.centroids = deque(maxlen=100)

        self.M_tens = torch.randn(
            1, self.model_config.latent_dim, self.model_config.latent_dim
        )
        self.centroids_tens = torch.randn(1, self.model_config.latent_dim)

        # define a starting metric (gamma_i = 0 & L = I_d)
        def G(z):
            return torch.inverse(
                (
                    torch.eye(self.latent_dim, device=z.device).unsqueeze(0)
                    * torch.exp(-torch.norm(z.unsqueeze(1), dim=-1) ** 2)
                    .unsqueeze(-1)
                    .unsqueeze(-1)
                ).sum(dim=1)
                + self.lbd * torch.eye(self.latent_dim).to(z.device)
            )

        def G_inv(z):
            return (
                torch.eye(self.latent_dim, device=z.device).unsqueeze(0)
                * torch.exp(-torch.norm(z.unsqueeze(1), dim=-1) ** 2)
                .unsqueeze(-1)
                .unsqueeze(-1)
            ).sum(dim=1) + self.lbd * torch.eye(self.latent_dim).to(z.device)

        self.G = G
        self.G_inv = G_inv

    def update(self):
        r"""
        As soon as the model has seen all the data points (i.e. at the end of 1 loop)
        we update the final metric function using \mu(x_i) as centroids
        """
        self._update_metric()

    def set_metric(self, metric: BaseMetric) -> None:
        r"""This method is called to set the metric network outputing the
        :math:`L_{\psi_i}` of the metric matrices

        Args:
            metric (BaseMetric): The metric module that need to be set to the model.

        """
        if not issubclass(type(metric), BaseMetric):
            raise BadInheritanceError(
                (
                    "Metric must inherit from BaseMetric class from "
                    "XGen.models.base_architectures.BaseMetric. Refer to documentation."
                )
            )

        self.metric = metric

    def forward(self, inputs: BaseDataset, **kwargs) -> ModelOutput:
        r"""
        The input data is first encoded. The reparametrization is used to produce a sample
        :math:`z_0` from the approximate posterior :math:`q_{\phi}(z|x)`. Then Riemannian
        Hamiltonian equations are solved using the generalized leapfrog integrator. In the meantime,
        the input data :math:`x` is fed to the metric network outputing the matrices
        :math:`L_{\psi}`. The metric is computed and used with the integrator.

        Args:
            inputs (BaseDataset): The training data with labels

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters
        """

        x = inputs["data"]

        encoder_output = self.encoder(x)
        mu, log_var = encoder_output.embedding, encoder_output.log_covariance

        std = torch.exp(0.5 * log_var)
        z0, eps0 = self._sample_gauss(mu, std)

        z = z0

        if self.training:
            # update the metric using batch data points
            L = self.metric(x)["L"]

            M = L @ torch.transpose(L, 1, 2)

            # store LL^T and mu(x_i) to update final metric
            self.M.append(M.detach().clone())
            self.centroids.append(mu.detach().clone())

            G_inv = (
                M.unsqueeze(0)
                * torch.exp(
                    -torch.norm(mu.unsqueeze(0) - z.unsqueeze(1), dim=-1) ** 2
                    / (self.temperature ** 2)
                )
                .unsqueeze(-1)
                .unsqueeze(-1)
            ).sum(dim=1) + self.lbd * torch.eye(self.latent_dim).to(x.device)

        else:
            G = self.G(z)
            G_inv = self.G_inv(z)
            L = torch.linalg.cholesky(G)

        G_log_det = -torch.logdet(G_inv)

        gamma = torch.randn_like(z0, device=x.device)
        rho = gamma / self.beta_zero_sqrt
        beta_sqrt_old = self.beta_zero_sqrt

        # sample \rho from N(0, G)
        rho = (L @ rho.unsqueeze(-1)).squeeze(-1)

        recon_x = self.decoder(z)["reconstruction"]

        for k in range(self.n_lf):

            # perform leapfrog steps

            # step 1
            rho_ = self._leap_step_1(recon_x, x, z, rho, G_inv, G_log_det)

            # step 2
            z = self._leap_step_2(recon_x, x, z, rho_, G_inv, G_log_det)

            recon_x = self.decoder(z)["reconstruction"]

            if self.training:

                G_inv = (
                    M.unsqueeze(0)
                    * torch.exp(
                        -torch.norm(mu.unsqueeze(0) - z.unsqueeze(1), dim=-1) ** 2
                        / (self.temperature ** 2)
                    )
                    .unsqueeze(-1)
                    .unsqueeze(-1)
                ).sum(dim=1) + self.lbd * torch.eye(self.latent_dim).to(x.device)

            else:
                # compute metric value on new z using final metric
                G = self.G(z)
                G_inv = self.G_inv(z)

            G_log_det = -torch.logdet(G_inv)

            # step 3
            rho__ = self._leap_step_3(recon_x, x, z, rho_, G_inv, G_log_det)

            # tempering
            beta_sqrt = self._tempering(k + 1, self.n_lf)
            rho = (beta_sqrt_old / beta_sqrt) * rho__
            beta_sqrt_old = beta_sqrt

        loss = self.loss_function(
            recon_x, x, z0, z, rho, eps0, gamma, mu, log_var, G_inv, G_log_det
        )

        output = ModelOutput(
            loss=loss,
            recon_x=recon_x,
            z=z,
            z0=z0,
            rho=rho,
            eps0=eps0,
            gamma=gamma,
            mu=mu,
            log_var=log_var,
            G_inv=G_inv,
            G_log_det=G_log_det,
        )

        return output

    def _leap_step_1(self, recon_x, x, z, rho, G_inv, G_log_det, steps=3):
        """
        Resolves first equation of generalized leapfrog integrator
        using fixed point iterations
        """

        def f_(rho_):
            H = self._hamiltonian(recon_x, x, z, rho_, G_inv, G_log_det)
            gz = grad(H, z, retain_graph=True)[0]
            return rho - 0.5 * self.eps_lf * gz

        rho_ = rho.clone()
        for _ in range(steps):
            rho_ = f_(rho_)
        return rho_

    def _leap_step_2(self, recon_x, x, z, rho, G_inv, G_log_det, steps=3):
        """
        Resolves second equation of generalized leapfrog integrator
        using fixed point iterations
        """
        H0 = self._hamiltonian(recon_x, x, z, rho, G_inv, G_log_det)
        grho_0 = grad(H0, rho)[0]

        def f_(z_):
            H = self._hamiltonian(recon_x, x, z_, rho, G_inv, G_log_det)
            grho = grad(H, rho, retain_graph=True)[0]
            return z + 0.5 * self.eps_lf * (grho_0 + grho)

        z_ = z.clone()
        for _ in range(steps):
            z_ = f_(z_)
        return z_

    def _leap_step_3(self, recon_x, x, z, rho, G_inv, G_log_det, steps=3):
        """
        Resolves third equation of generalized leapfrog integrator
        """
        H = self._hamiltonian(recon_x, x, z, rho, G_inv, G_log_det)
        gz = grad(H, z, create_graph=True)[0]
        return rho - 0.5 * self.eps_lf * gz

    def _hamiltonian(self, recon_x, x, z, rho, G_inv=None, G_log_det=None):
        """
        Computes the Hamiltonian function.
        used for RHVAE
        """
        norm = (
            torch.transpose(rho.unsqueeze(-1), 1, 2) @ G_inv @ rho.unsqueeze(-1)
        ).sum()

        return -self._log_p_xz(recon_x, x, z).sum() + 0.5 * norm + 0.5 * G_log_det.sum()

    def _update_metric(self):
        # convert to 1 big tensor

        self.M_tens = torch.cat(list(self.M))
        self.centroids_tens = torch.cat(list(self.centroids))

        # define new metric
        def G(z):
            return torch.inverse(
                (
                    self.M_tens.unsqueeze(0)
                    * torch.exp(
                        -torch.norm(
                            self.centroids_tens.unsqueeze(0) - z.unsqueeze(1), dim=-1
                        )
                        ** 2
                        / (self.temperature ** 2)
                    )
                    .unsqueeze(-1)
                    .unsqueeze(-1)
                ).sum(dim=1)
                + self.lbd * torch.eye(self.latent_dim).to(z.device)
            )

        def G_inv(z):
            return (
                self.M_tens.unsqueeze(0)
                * torch.exp(
                    -torch.norm(
                        self.centroids_tens.unsqueeze(0) - z.unsqueeze(1), dim=-1
                    )
                    ** 2
                    / (self.temperature ** 2)
                )
                .unsqueeze(-1)
                .unsqueeze(-1)
            ).sum(dim=1) + self.lbd * torch.eye(self.latent_dim).to(z.device)

        self.G = G
        self.G_inv = G_inv
        self.M = deque(maxlen=100)
        self.centroids = deque(maxlen=100)

    def loss_function(
        self, recon_x, x, z0, zK, rhoK, eps0, gamma, mu, log_var, G_inv, G_log_det
    ):

        logpxz = self._log_p_xz(recon_x, x, zK)  # log p(x, z_K)
        logrhoK = (
            -0.5
            * (torch.transpose(rhoK.unsqueeze(-1), 1, 2) @ G_inv @ rhoK.unsqueeze(-1))
            .squeeze()
            .squeeze()
            - 0.5 * G_log_det
            # - torch.log(torch.tensor([2 * np.pi]).to(x.device)) * self.latent_dim / 2
        )  # log p(\rho_K)

        logp = logpxz + logrhoK

        # define a N(0, I) distribution
        normal = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.latent_dim).to(x.device),
            covariance_matrix=torch.eye(self.latent_dim).to(x.device),
        )

        logq = normal.log_prob(eps0) - 0.5 * log_var.sum(dim=1)  # log(q(z_0|x))

        return -(logp - logq).mean(dim=0)

    def _sample_gauss(self, mu, std):
        # Reparametrization trick
        # Sample N(0, I)
        eps = torch.randn_like(std)
        return mu + eps * std, eps

    def _tempering(self, k, K):
        """Perform tempering step"""

        beta_k = (
            (1 - 1 / self.beta_zero_sqrt) * (k / K) ** 2
        ) + 1 / self.beta_zero_sqrt

        return 1 / beta_k

    def _log_p_x_given_z(self, recon_x, x):

        if self.model_config.reconstruction_loss == "mse":
            # sigma is taken as I_D
            recon_loss = (
                -0.5
                * F.mse_loss(
                    recon_x.reshape(x.shape[0], -1),
                    x.reshape(x.shape[0], -1),
                    reduction="none",
                ).sum(dim=-1)
            )
            -torch.log(torch.tensor([2 * np.pi]).to(x.device)) * np.prod(
                self.input_dim
            ) / 2

        elif self.model_config.reconstruction_loss == "bce":

            recon_loss = -F.binary_cross_entropy(
                recon_x.reshape(x.shape[0], -1),
                x.reshape(x.shape[0], -1),
                reduction="none",
            ).sum(dim=-1)

        return recon_loss

    def _log_z(self, z):
        """
        Return Normal density function as prior on z
        """

        # define a N(0, I) distribution
        normal = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.latent_dim).to(z.device),
            covariance_matrix=torch.eye(self.latent_dim).to(z.device),
        )
        return normal.log_prob(z)

    def _log_p_xz(self, recon_x, x, z):
        """
        Estimate log(p(x, z)) using Bayes rule
        """
        logpxz = self._log_p_x_given_z(recon_x, x)
        logpz = self._log_z(z)
        return logpxz + logpz

    def get_nll(self, data, n_samples=1, batch_size=100):
        """
        Function computed the estimate negative log-likelihood of the model. It uses importance
        sampling method with the approximate posterior distribution. This may take a while.

        Args:
            data (torch.Tensor): The input data from which the log-likelihood should be estimated.
                Data must be of shape [Batch x n_channels x ...]
            n_samples (int): The number of importance samples to use for estimation
            batch_size (int): The batchsize to use to avoid memory issues
        """
        normal = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.model_config.latent_dim).to(data.device),
            covariance_matrix=torch.eye(self.model_config.latent_dim).to(data.device),
        )

        if n_samples <= batch_size:
            n_full_batch = 1
        else:
            n_full_batch = n_samples // batch_size
            n_samples = batch_size

        log_p = []

        for i in range(len(data)):
            x = data[i].unsqueeze(0)

            log_p_x = []

            for j in range(n_full_batch):

                x_rep = torch.cat(batch_size * [x])

                encoder_output = self.encoder(x_rep)
                mu, log_var = encoder_output.embedding, encoder_output.log_covariance

                std = torch.exp(0.5 * log_var)
                z0, eps = self._sample_gauss(mu, std)
                gamma = torch.randn_like(z0, device=x.device)
                rho = gamma / self.beta_zero_sqrt

                z = z0
                beta_sqrt_old = self.beta_zero_sqrt

                G = self.G(z0)
                G_inv = self.G_inv(z0)
                G_log_det = -torch.logdet(G_inv)
                L = torch.linalg.cholesky(G)

                # initialization
                gamma = torch.randn_like(z0, device=z.device)
                rho = gamma / self.beta_zero_sqrt
                beta_sqrt_old = self.beta_zero_sqrt

                rho = (L @ rho.unsqueeze(-1)).squeeze(
                    -1
                )  # sample from the multivariate N(0, G)

                recon_x = self.decoder(z)["reconstruction"]

                for k in range(self.n_lf):

                    # perform leapfrog steps

                    # step 1
                    rho_ = self._leap_step_1(recon_x, x_rep, z, rho, G_inv, G_log_det)

                    # step 2
                    z = self._leap_step_2(recon_x, x_rep, z, rho_, G_inv, G_log_det)

                    recon_x = self.decoder(z)["reconstruction"]

                    G_inv = self.G_inv(z)
                    G_log_det = -torch.logdet(G_inv)

                    # step 3
                    rho__ = self._leap_step_3(recon_x, x_rep, z, rho_, G_inv, G_log_det)

                    # tempering steps
                    beta_sqrt = self._tempering(k + 1, self.n_lf)
                    rho = (beta_sqrt_old / beta_sqrt) * rho__
                    beta_sqrt_old = beta_sqrt

                log_q_z0_given_x = -0.5 * (
                    log_var + (z0 - mu) ** 2 / torch.exp(log_var)
                ).sum(dim=-1)
                log_p_z = -0.5 * (z ** 2).sum(dim=-1)

                log_p_rho0 = normal.log_prob(gamma) - torch.logdet(
                    L / self.beta_zero_sqrt
                )  # rho0 ~ N(0, 1/beta_0 * G(z0))

                log_p_rho = (
                    -0.5
                    * (
                        torch.transpose(rho.unsqueeze(-1), 1, 2)
                        @ G_inv
                        @ rho.unsqueeze(-1)
                    )
                    .squeeze()
                    .squeeze()
                    - 0.5 * G_log_det
                ) - torch.log(
                    torch.tensor([2 * np.pi]).to(z.device)
                ) * self.latent_dim / 2  # rho0 ~ N(0, G(z))

                if self.model_config.reconstruction_loss == "mse":

                    log_p_x_given_z = -0.5 * F.mse_loss(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1) - torch.tensor(
                        [np.prod(self.input_dim) / 2 * np.log(np.pi * 2)]
                    ).to(
                        data.device
                    )  # decoding distribution is assumed unit variance  N(mu, I)

                elif self.model_config.reconstruction_loss == "bce":

                    log_p_x_given_z = -F.binary_cross_entropy(
                        recon_x.reshape(x_rep.shape[0], -1),
                        x_rep.reshape(x_rep.shape[0], -1),
                        reduction="none",
                    ).sum(dim=-1)

                log_p_x.append(
                    log_p_x_given_z
                    + log_p_z
                    + log_p_rho
                    - log_p_rho0
                    - log_q_z0_given_x
                )  # N*log(2*pi) simplifies in prior and posterior

            log_p_x = torch.cat(log_p_x)

            log_p.append((torch.logsumexp(log_p_x, 0) - np.log(len(log_p_x))).item())

        return np.mean(log_p)

    def save(self, dir_path: str):
        """Method to save the model at a specific location

        Args:
            dir_path (str): The path where the model should be saved. If the path
                path does not exist a folder will be created at the provided location.
        """

        # This creates the dir if not available
        super().save(dir_path)

        model_path = dir_path

        model_dict = {
            "M": deepcopy(self.M_tens.clone().detach()),
            "centroids": deepcopy(self.centroids_tens.clone().detach()),
            "model_state_dict": deepcopy(self.state_dict()),
        }

        if not self.model_config.uses_default_metric:
            with open(os.path.join(model_path, "metric.pkl"), "wb") as fp:
                cloudpickle.register_pickle_by_value(inspect.getmodule(self.metric))
                cloudpickle.dump(self.metric, fp)

        torch.save(model_dict, os.path.join(model_path, "model.pt"))

    @classmethod
    def _load_custom_metric_from_folder(cls, dir_path):

        file_list = os.listdir(dir_path)
        cls._check_python_version_from_folder(dir_path=dir_path)

        if "metric.pkl" not in file_list:
            raise FileNotFoundError(
                f"Missing metric pkl file ('metric.pkl') in"
                f"{dir_path}... This file is needed to rebuild custom metrics."
                " Cannot perform model building."
            )

        else:
            with open(os.path.join(dir_path, "metric.pkl"), "rb") as fp:
                metric = CPU_Unpickler(fp).load()

        return metric

    @classmethod
    def _load_metric_matrices_and_centroids(cls, dir_path):
        """this function can be called safely since it is called after
        _load_model_weights_from_folder which handles FileNotFoundError and
        loading issues"""

        path_to_model_weights = os.path.join(dir_path, "model.pt")

        model_weights = torch.load(path_to_model_weights, map_location="cpu")

        if "M" not in model_weights.keys():
            raise KeyError(
                "Metric M matrices are not available in 'model.pt' file. Got keys:"
                f"{model_weights.keys()}. These are needed to build the metric."
            )

        metric_M = model_weights["M"]

        if "centroids" not in model_weights.keys():
            raise KeyError(
                "Metric centroids are not available in 'model.pt' file. Got keys:"
                f"{model_weights.keys()}. These are needed to build the metric."
            )

        metric_centroids = model_weights["centroids"]

        return metric_M, metric_centroids

    @classmethod
    def load_from_folder(cls, dir_path):
        """Class method to be used to load the model from a specific folder

        Args:
            dir_path (str): The path where the model should have been be saved.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoder.pkl`` (resp.
                ``decoder.pkl`` or/and ``metric.pkl``) if a custom encoder (resp. decoder or/and
                metric) was provided
        """

        model_config = cls._load_model_config_from_folder(dir_path)
        model_weights = cls._load_model_weights_from_folder(dir_path)

        if not model_config.uses_default_encoder:
            encoder = cls._load_custom_encoder_from_folder(dir_path)

        else:
            encoder = None

        if not model_config.uses_default_decoder:
            decoder = cls._load_custom_decoder_from_folder(dir_path)

        else:
            decoder = None

        if not model_config.uses_default_metric:
            metric = cls._load_custom_metric_from_folder(dir_path)

        else:
            metric = None

        model = cls(model_config, encoder=encoder, decoder=decoder, metric=metric)

        metric_M, metric_centroids = cls._load_metric_matrices_and_centroids(dir_path)

        model.M_tens = metric_M
        model.centroids_tens = metric_centroids

        model.G = create_metric(model)
        model.G_inv = create_inverse_metric(model)

        model.load_state_dict(model_weights)

        return model

    @classmethod
    def load_from_hf_hub(
        cls, hf_hub_path: str, allow_pickle: bool = False
    ):  # pragma: no cover
        """Class method to be used to load a pretrained model from the Hugging Face hub

        Args:
            hf_hub_path (str): The path where the model should have been be saved on the
                hugginface hub.

        .. note::
            This function requires the folder to contain:

            - | a ``model_config.json`` and a ``model.pt`` if no custom architectures were provided

            **or**

            - | a ``model_config.json``, a ``model.pt`` and a ``encoder.pkl`` (resp.
                ``decoder.pkl`` and ``metric.pkl``) if a custom encoder (resp. decoder and/or
                metric) was provided
        """

        if not hf_hub_is_available():
            raise ModuleNotFoundError(
                "`huggingface_hub` package must be installed to load models from the HF hub. "
                "Run `python -m pip install huggingface_hub` and log in to your account with "
                "`huggingface-cli login`."
            )

        else:
            from huggingface_hub import hf_hub_download

        logger.info(f"Downloading {cls.__name__} files for rebuilding...")

        config_path = hf_hub_download(repo_id=hf_hub_path, filename="model_config.json")
        dir_path = os.path.dirname(config_path)

        _ = hf_hub_download(repo_id=hf_hub_path, filename="model.pt")

        model_config = cls._load_model_config_from_folder(dir_path)

        if (
            cls.__name__ + "Config" != model_config.name
            and cls.__name__ + "_Config" != model_config.name
        ):
            warnings.warn(
                f"You are trying to load a "
                f"`{ cls.__name__}` while a "
                f"`{model_config.name}` is given."
            )

        model_weights = cls._load_model_weights_from_folder(dir_path)

        if (
            not model_config.uses_default_encoder
            or not model_config.uses_default_decoder
            or not model_config.uses_default_metric
        ) and not allow_pickle:
            warnings.warn(
                "You are about to download pickled files from the HF hub that may have "
                "been created by a third party and so could potentially harm your computer. If you "
                "are sure that you want to download them set `allow_pickle=true`."
            )

        else:

            if not model_config.uses_default_encoder:
                _ = hf_hub_download(repo_id=hf_hub_path, filename="encoder.pkl")
                encoder = cls._load_custom_encoder_from_folder(dir_path)

            else:
                encoder = None

            if not model_config.uses_default_decoder:
                _ = hf_hub_download(repo_id=hf_hub_path, filename="decoder.pkl")
                decoder = cls._load_custom_decoder_from_folder(dir_path)

            else:
                decoder = None

            if not model_config.uses_default_metric:
                _ = hf_hub_download(repo_id=hf_hub_path, filename="metric.pkl")
                metric = cls._load_custom_metric_from_folder(dir_path)

            else:
                metric = None

            logger.info(f"Successfully downloaded {cls.__name__} model!")

            model = cls(model_config, encoder=encoder, decoder=decoder, metric=metric)

            metric_M, metric_centroids = cls._load_metric_matrices_and_centroids(
                dir_path
            )

            model.M_tens = metric_M
            model.centroids_tens = metric_centroids

            model.G = create_metric(model)
            model.G_inv = create_inverse_metric(model)

            model.load_state_dict(model_weights)

            return model



"""
In this module are implemented so normalizing flows that can be used within the vae model or for 
sampling.

By convention, each implemented model is stored in a folder located in 
:class:`XGen.models.normalizing_flows` and named likewise the model. The following modules can be 
found in this folder:

- | *modelname_config.py*: Contains a :class:`ModelNameConfig` instance inheriting
    from :class:`~XGen.models.normalizing_flows.BaseNF`. 
- | *modelname_model.py*: An implementation of the model inheriting either from
    :class:`~XGen.models.normalizing_flows.BaseNFConfig` 
- *modelname_utils.py* (optional): A module where utils methods are stored.

All the codes are inspired from:

- | https://github.com/kamenbliznashki/normalizing_flows
- | https://github.com/karpathy/pytorch-normalizing-flows
- | https://github.com/ikostrikov/pytorch-flows
"""

from .base import BaseNF, BaseNFConfig, NFModel
from .iaf import IAF, IAFConfig
from .made import MADE, MADEConfig
from .maf import MAF, MAFConfig
from .pixelcnn import PixelCNN, PixelCNNConfig
from .planar_flow import PlanarFlow, PlanarFlowConfig
from .radial_flow import RadialFlow, RadialFlowConfig

__all__ = [
    "BaseNF",
    "BaseNFConfig",
    "NFModel",
    "MADE",
    "MADEConfig",
    "MAF",
    "MAFConfig",
    "IAF",
    "IAFConfig",
    "PlanarFlow",
    "PlanarFlowConfig",
    "RadialFlow",
    "RadialFlowConfig",
    "PixelCNN",
    "PixelCNNConfig",
]



from pydantic.dataclasses import dataclass

from ..base import BaseSamplerConfig


@dataclass
class RHVAESamplerConfig(BaseSamplerConfig):
    """RHVAESampler config class.

    Parameters:
        num_samples (int): The number of samples to generate. Default: 1
        batch_size (int): The number of samples per batch. Batching is used to speed up
            generation and avoid memory overflows. Default: 50
        mcmc_steps (int): The number of MCMC steps to use in the latent space HMC sampler.
            Default: 100
        n_lf (int): The number of leapfrog to use in the integrator of the HMC sampler.
            Default: 15
        eps_lf (float): The leapfrog stepsize in the integrator of the HMC sampler. Default: 3e-2
        random_start (bool): Initialization of the latent space sampler. If False, the sampler
            starts the Markov chain on the metric centroids. If True , a random start is applied.
            Default: False
    """

    mcmc_steps_nbr: int = 100
    n_lf: int = 15
    eps_lf: float = 0.03
    beta_zero: float = 1.0



"""Implementation of a Vanilla Autoencoder model.

Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:

"""

from .ae_config import AEConfig
from .ae_model import AE

__all__ = ["AE", "AEConfig"]



from pydantic.dataclasses import dataclass

from ..ae import AEConfig


@dataclass
class VQVAEConfig(AEConfig):
    r"""
    Vector Quantized VAE model config config class

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        commitment_loss_factor (float): The commitment loss factor in the loss. Default: 0.25.
        quantization_loss_factor: The quantization loss factor in the loss. Default: 1.
        num_embedding (int): The number of embedding points. Default: 512
        use_ema (bool): Whether to use the Exponential Movng Average Update (EMA). Default: False.
        decay (float): The decay to apply in the EMA update. Must be in [0, 1]. Default: 0.99.
    """
    commitment_loss_factor: float = 0.25
    quantization_loss_factor: float = 1.0
    num_embeddings: int = 512
    use_ema: bool = False
    decay: float = 0.99

    def __post_init_post_parse__(self):
        if self.use_ema:
            assert 0 <= self.decay <= 1, (
                "The decay in the EMA update must be in [0, 1]. " f"Got {self.decay}."
            )



from typing import Union

import torch
import torch.nn as nn
import torch.nn.functional as F


class MaskedConv2d(nn.Conv2d):
    def __init__(
        self,
        mask_type: str,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: Union[str, int] = 0,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
            device,
            dtype,
        )

        self.register_buffer("mask", torch.ones_like(self.weight))

        _, _, kH, kW = self.weight.shape

        if mask_type == "A":
            self.mask[:, :, kH // 2, kW // 2 :] = 0
            self.mask[:, :, kH // 2 + 1 :] = 0

        else:
            self.mask[:, :, kH // 2, kW // 2 + 1 :] = 0
            self.mask[:, :, kH // 2 + 1 :] = 0

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # self.weight.data *= self.mask

        return F.conv2d(
            input,
            self.weight * self.mask,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
        )



from pydantic.dataclasses import dataclass

from XGen.config import BaseConfig


@dataclass
class AutoConfig(BaseConfig):
    @classmethod
    def from_json_file(cls, json_path):
        """Creates a :class:`~XGen.config.BaseAEConfig` instance from a JSON config file. It
        builds automatically the correct config for any `XGen.models`.

        Args:
            json_path (str): The path to the json file containing all the parameters

        Returns:
            :class:`BaseAEConfig`: The created instance
        """

        config_dict = cls._dict_from_json(json_path)
        config_name = config_dict.pop("name")

        if config_name == "BaseAEConfig":
            from ..base import BaseAEConfig

            model_config = BaseAEConfig.from_json_file(json_path)

        elif config_name == "Adversarial_AE_Config":
            from ..adversarial_ae import Adversarial_AE_Config

            model_config = Adversarial_AE_Config.from_json_file(json_path)

        elif config_name == "AEConfig":
            from ..ae import AEConfig

            model_config = AEConfig.from_json_file(json_path)

        elif config_name == "BetaTCVAEConfig":
            from ..beta_tc_vae import BetaTCVAEConfig

            model_config = BetaTCVAEConfig.from_json_file(json_path)

        elif config_name == "BetaVAEConfig":
            from ..beta_vae import BetaVAEConfig

            model_config = BetaVAEConfig.from_json_file(json_path)

        elif config_name == "DisentangledBetaVAEConfig":
            from ..disentangled_beta_vae import DisentangledBetaVAEConfig

            model_config = DisentangledBetaVAEConfig.from_json_file(json_path)

        elif config_name == "FactorVAEConfig":
            from ..factor_vae import FactorVAEConfig

            model_config = FactorVAEConfig.from_json_file(json_path)

        elif config_name == "HVAEConfig":
            from ..hvae import HVAEConfig

            model_config = HVAEConfig.from_json_file(json_path)

        elif config_name == "INFOVAE_MMD_Config":
            from ..info_vae import INFOVAE_MMD_Config

            model_config = INFOVAE_MMD_Config.from_json_file(json_path)

        elif config_name == "IWAEConfig":
            from ..iwae import IWAEConfig

            model_config = IWAEConfig.from_json_file(json_path)

        elif config_name == "MSSSIM_VAEConfig":
            from ..msssim_vae import MSSSIM_VAEConfig

            model_config = MSSSIM_VAEConfig.from_json_file(json_path)

        elif config_name == "RAE_GP_Config":
            from ..rae_gp import RAE_GP_Config

            model_config = RAE_GP_Config.from_json_file(json_path)

        elif config_name == "RAE_L2_Config":
            from ..rae_l2 import RAE_L2_Config

            model_config = RAE_L2_Config.from_json_file(json_path)

        elif config_name == "RHVAEConfig":
            from ..rhvae import RHVAEConfig

            model_config = RHVAEConfig.from_json_file(json_path)

        elif config_name == "SVAEConfig":
            from ..svae import SVAEConfig

            model_config = SVAEConfig.from_json_file(json_path)

        elif config_name == "VAEConfig":
            from ..vae import VAEConfig

            model_config = VAEConfig.from_json_file(json_path)

        elif config_name == "VAEGANConfig":
            from ..vae_gan import VAEGANConfig

            model_config = VAEGANConfig.from_json_file(json_path)

        elif config_name == "VAE_IAF_Config":
            from ..vae_iaf import VAE_IAF_Config

            model_config = VAE_IAF_Config.from_json_file(json_path)

        elif config_name == "VAE_LinNF_Config":
            from ..vae_lin_nf import VAE_LinNF_Config

            model_config = VAE_LinNF_Config.from_json_file(json_path)

        elif config_name == "VAMPConfig":
            from ..vamp import VAMPConfig

            model_config = VAMPConfig.from_json_file(json_path)

        elif config_name == "VQVAEConfig":
            from ..vq_vae import VQVAEConfig

            model_config = VQVAEConfig.from_json_file(json_path)

        elif config_name == "WAE_MMD_Config":
            from ..wae_mmd import WAE_MMD_Config

            model_config = WAE_MMD_Config.from_json_file(json_path)

        elif config_name == "MAFConfig":
            from ..normalizing_flows import MAFConfig

            model_config = MAFConfig.from_json_file(json_path)

        elif config_name == "IAFConfig":
            from ..normalizing_flows import IAFConfig

            model_config = IAFConfig.from_json_file(json_path)

        elif config_name == "PlanarFlowConfig":
            from ..normalizing_flows import PlanarFlowConfig

            model_config = PlanarFlowConfig.from_json_file(json_path)

        elif config_name == "RadialFlowConfig":
            from ..normalizing_flows import RadialFlowConfig

            model_config = RadialFlowConfig.from_json_file(json_path)

        elif config_name == "MADEConfig":
            from ..normalizing_flows import MADEConfig

            model_config = MADEConfig.from_json_file(json_path)

        elif config_name == "PixelCNNConfig":
            from ..normalizing_flows import PixelCNNConfig

            model_config = PixelCNNConfig.from_json_file(json_path)

        elif config_name == "PoincareVAEConfig":
            from ..pvae import PoincareVAEConfig

            model_config = PoincareVAEConfig.from_json_file(json_path)

        elif config_name == "CIWAEConfig":
            from ..ciwae import CIWAEConfig

            model_config = CIWAEConfig.from_json_file(json_path)

        elif config_name == "MIWAEConfig":
            from ..miwae import MIWAEConfig

            model_config = MIWAEConfig.from_json_file(json_path)

        elif config_name == "PIWAEConfig":
            from ..piwae import PIWAEConfig

            model_config = PIWAEConfig.from_json_file(json_path)

        else:
            raise NameError(
                "Cannot reload automatically the model configuration... "
                f"The model name in the `model_config.json may be corrupted. Got `{config_name}`"
            )

        return model_config



"""This module is the implementation of the Multiply Importance Weighted Autoencoder
proposed in (https://arxiv.org/abs/1802.04537).

Available samplers
-------------------

.. autosummary::
    ~XGen.samplers.NormalSampler
    ~XGen.samplers.GaussianMixtureSampler
    ~XGen.samplers.TwoStageVAESampler
    ~XGen.samplers.MAFSampler
    ~XGen.samplers.IAFSampler
    :nosignatures:

"""

from .miwae_config import MIWAEConfig
from .miwae_model import MIWAE

__all__ = ["MIWAE", "MIWAEConfig"]



import os
from typing import Optional

import torch
import torch.nn.functional as F

from ...data.datasets import BaseDataset
from ..ae import AE
from ..base.base_utils import ModelOutput
from ..nn import BaseDecoder, BaseEncoder
from .vq_vae_config import VQVAEConfig
from .vq_vae_utils import Quantizer, QuantizerEMA


class VQVAE(AE):
    r"""
    Vector Quantized-VAE model.

    Args:
        model_config (VQVAEConfig): The Variational Autoencoder configuration setting the main
            parameters of the model.

        encoder (BaseEncoder): An instance of BaseEncoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

        decoder (BaseDecoder): An instance of BaseDecoder (inheriting from `torch.nn.Module` which
            plays the role of encoder. This argument allows you to use your own neural networks
            architectures if desired. If None is provided, a simple Multi Layer Preception
            (https://en.wikipedia.org/wiki/Multilayer_perceptron) is used. Default: None.

    .. note::
        For high dimensional data we advice you to provide you own network architectures. With the
        provided MLP you may end up with a ``MemoryError``.
    """

    def __init__(
        self,
        model_config: VQVAEConfig,
        encoder: Optional[BaseEncoder] = None,
        decoder: Optional[BaseDecoder] = None,
    ):

        AE.__init__(self, model_config=model_config, encoder=encoder, decoder=decoder)

        self._set_quantizer(model_config)

        self.model_name = "VQVAE"

    def _set_quantizer(self, model_config):

        if model_config.input_dim is None:
            raise AttributeError(
                "No input dimension provided !"
                "'input_dim' parameter of VQVAEConfig instance must be set to 'data_shape' where "
                "the shape of the data is (C, H, W ..). Unable to set quantizer."
            )

        x = torch.randn((2,) + self.model_config.input_dim)
        z = self.encoder(x).embedding
        if len(z.shape) == 2:
            z = z.reshape(z.shape[0], 1, 1, -1)

        z = z.permute(0, 2, 3, 1)

        self.model_config.embedding_dim = z.shape[-1]
        if model_config.use_ema:
            self.quantizer = QuantizerEMA(model_config=model_config)

        else:
            self.quantizer = Quantizer(model_config=model_config)

    def forward(self, inputs: BaseDataset, **kwargs):
        """
        The VAE model

        Args:
            inputs (BaseDataset): The training dataset with labels

        Returns:
            ModelOutput: An instance of ModelOutput containing all the relevant parameters

        """

        x = inputs["data"]
        uses_ddp = kwargs.pop("uses_ddp", False)

        encoder_output = self.encoder(x)

        embeddings = encoder_output.embedding

        reshape_for_decoding = False

        if len(embeddings.shape) == 2:
            embeddings = embeddings.reshape(embeddings.shape[0], 1, 1, -1)
            reshape_for_decoding = True

        embeddings = embeddings.permute(0, 2, 3, 1)

        quantizer_output = self.quantizer(embeddings, uses_ddp=uses_ddp)

        quantized_embed = quantizer_output.quantized_vector
        quantized_indices = quantizer_output.quantized_indices

        if reshape_for_decoding:
            quantized_embed = quantized_embed.reshape(embeddings.shape[0], -1)

        recon_x = self.decoder(quantized_embed).reconstruction

        loss, recon_loss, vq_loss = self.loss_function(recon_x, x, quantizer_output)

        output = ModelOutput(
            recon_loss=recon_loss,
            vq_loss=vq_loss,
            loss=loss,
            recon_x=recon_x,
            z=quantized_embed,
            quantized_indices=quantized_indices,
        )

        return output

    def loss_function(self, recon_x, x, quantizer_output):

        recon_loss = F.mse_loss(
            recon_x.reshape(x.shape[0], -1), x.reshape(x.shape[0], -1), reduction="none"
        ).sum(dim=-1)

        vq_loss = quantizer_output.loss

        return (
            (recon_loss + vq_loss).mean(dim=0),
            recon_loss.mean(dim=0),
            vq_loss.mean(dim=0),
        )

    def _sample_gauss(self, mu, std):
        # Reparametrization trick
        # Sample N(0, I)
        eps = torch.randn_like(std)
        return mu + eps * std, eps



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class BetaTCVAEConfig(VAEConfig):
    r"""
    :math:`\beta`-TCVAE model config config class

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        alpha (float): The balancing factor before the Index code Mutual Info. Default: 1
        beta (float): The balancing factor before the Total Correlation. Default: 1
        gamma (float): The balancing factor before the dimension-wise KL. Default: 1
        use_mss (bool): Use Minibatch Stratified Sampling. If False: uses Minibatch Weighted
            Sampling. Default: True
    """
    alpha: float = 1.0
    beta: float = 1.0
    gamma: float = 1.0
    use_mss: bool = True



import torch
import torch.nn as nn


class FactorVAEDiscriminator(nn.Module):
    def __init__(self, latent_dim=16, hidden_units=1000) -> None:

        nn.Module.__init__(self)

        self.layers = nn.Sequential(
            nn.Linear(latent_dim, hidden_units),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_units, hidden_units),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_units, hidden_units),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_units, hidden_units),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_units, hidden_units),
            nn.LeakyReLU(0.2),
            nn.Linear(hidden_units, 2),
        )

    def forward(self, z: torch.Tensor):
        return self.layers(z)



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class MSSSIM_VAEConfig(VAEConfig):
    r"""
    VAE using using perceptual similarity metrics (MS-SSIM) model config config class

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        beta (float): The balancing factor. Default: 1
    """

    beta: float = 1.0
    window_size: int = 11



"""This module implements the dual optimizer trainer using two distinct optimizers for the encoder 
and the decoder. It is suitable for all models but must be used in particular to train a 
:class:`~XGen.models.RAE_L2`.

Available models:
------------------

.. autosummary::
    ~XGen.models.AE
    ~XGen.models.VAE
    ~XGen.models.BetaVAE
    ~XGen.models.DisentangledBetaVAE
    ~XGen.models.BetaTCVAE
    ~XGen.models.IWAE
    ~XGen.models.MSSSIM_VAE
    ~XGen.models.INFOVAE_MMD
    ~XGen.models.WAE_MMD
    ~XGen.models.VAMP
    ~XGen.models.SVAE
    ~XGen.models.VQVAE
    ~XGen.models.RAE_GP
    ~XGen.models.RAE_L2
    ~XGen.models.HVAE
    ~XGen.models.RHVAE
    :nosignatures:
"""

from .coupled_optimizer_trainer import CoupledOptimizerTrainer
from .coupled_optimizer_trainer_config import CoupledOptimizerTrainerConfig

__all__ = ["CoupledOptimizerTrainer", "CoupledOptimizerTrainerConfig"]



""" 
This is the heart of XGen! 
Here are implemented some of the most common (Variational) Autoencoders models.

By convention, each implemented model is stored in a folder located in :class:`XGen.models`
and named likewise the model. The following modules can be found in this folder:

- | *modelname_config.py*: Contains a :class:`ModelNameConfig` instance inheriting
    from either :class:`~XGen.models.base.AEConfig` for Autoencoder models or 
    :class:`~XGen.models.base.VAEConfig` for Variational Autoencoder models. 
- | *modelname_model.py*: An implementation of the model inheriting either from
    :class:`~XGen.models.AE` for Autoencoder models or 
    :class:`~XGen.models.base.VAE` for Variational Autoencoder models. 
- *modelname_utils.py* (optional): A module where utils methods are stored.
"""

from .adversarial_ae import Adversarial_AE, Adversarial_AE_Config
from .ae import AE, AEConfig
from .auto_model import AutoModel
from .base import BaseAE, BaseAEConfig
from .beta_tc_vae import BetaTCVAE, BetaTCVAEConfig
from .beta_vae import BetaVAE, BetaVAEConfig
from .ciwae import CIWAE, CIWAEConfig
from .disentangled_beta_vae import DisentangledBetaVAE, DisentangledBetaVAEConfig
from .factor_vae import FactorVAE, FactorVAEConfig
from .hvae import HVAE, HVAEConfig
from .info_vae import INFOVAE_MMD, INFOVAE_MMD_Config
from .iwae import IWAE, IWAEConfig
from .miwae import MIWAE, MIWAEConfig
from .msssim_vae import MSSSIM_VAE, MSSSIM_VAEConfig
from .piwae import PIWAE, PIWAEConfig
from .pvae import PoincareVAE, PoincareVAEConfig
from .rae_gp import RAE_GP, RAE_GP_Config
from .rae_l2 import RAE_L2, RAE_L2_Config
from .rhvae import RHVAE, RHVAEConfig
from .svae import SVAE, SVAEConfig
from .vae import VAE, VAEConfig
from .vae_gan import VAEGAN, VAEGANConfig
from .vae_iaf import VAE_IAF, VAE_IAF_Config
from .vae_lin_nf import VAE_LinNF, VAE_LinNF_Config
from .vamp import VAMP, VAMPConfig
from .vq_vae import VQVAE, VQVAEConfig
from .wae_mmd import WAE_MMD, WAE_MMD_Config

__all__ = [
    "AutoModel",
    "BaseAE",
    "BaseAEConfig",
    "AE",
    "AEConfig",
    "WAE_MMD",
    "WAE_MMD_Config",
    "VAE",
    "VAEConfig",
    "BetaVAE",
    "BetaVAEConfig",
    "VAMP",
    "VAMPConfig",
    "HVAE",
    "HVAEConfig",
    "RHVAE",
    "RHVAEConfig",
    "RAE_L2",
    "RAE_L2_Config",
    "RAE_GP",
    "RAE_GP_Config",
    "IWAE",
    "IWAEConfig",
    "INFOVAE_MMD",
    "INFOVAE_MMD_Config",
    "VQVAE",
    "VQVAEConfig",
    "Adversarial_AE",
    "Adversarial_AE_Config",
    "VAEGAN",
    "VAEGANConfig",
    "MSSSIM_VAE",
    "MSSSIM_VAEConfig",
    "SVAE",
    "SVAEConfig",
    "DisentangledBetaVAE",
    "DisentangledBetaVAEConfig",
    "FactorVAE",
    "FactorVAEConfig",
    "BetaTCVAE",
    "BetaTCVAEConfig",
    "VAE_LinNF",
    "VAE_LinNF_Config",
    "VAE_IAF",
    "VAE_IAF_Config",
    "PoincareVAE",
    "PoincareVAEConfig",
    "CIWAE",
    "CIWAEConfig",
    "MIWAE",
    "MIWAEConfig",
    "PIWAE",
    "PIWAEConfig",
]



from pydantic.dataclasses import dataclass

from ..vae import VAEConfig


@dataclass
class MIWAEConfig(VAEConfig):
    """Multiply IWAE model config class.

    Parameters:
        input_dim (tuple): The input_data dimension.
        latent_dim (int): The latent space dimension. Default: None.
        reconstruction_loss (str): The reconstruction loss to use ['bce', 'mse']. Default: 'mse'
        number_gradient_estimates (int): Number of (M-)estimates to use for the gradient
            estimate. Default: 5
        number_samples (int): Number of samples to use on the Monte-Carlo estimation. Default: 10
    """

    number_gradient_estimates: int = 5
    number_samples: int = 10
